"""
Module to parse Master Boot Record partition tables and return a (reasonably)
sane object.
"""
# TODO: add EBR support
# TODO: add comparisson operators for the classes?
#       this just came up after a small refactoring, and it'd be a simpler way
#       to test things
import struct


class Partition:
    """Handles one entry in the MBR table."""
    partition_types = {
        0x00: "Empty",
        0x01: "FAT12 CHS",
        0x04: "FAT16 CHS",
        0x05: "Microsoft Extended",
        0x06: "FAT16 CHS 32MB",
        0x07: "NTFS",
        0x0b: "FAT32 CHS",
        0x0c: "FAT32 LBA",
        0x0e: "FAT16 LBA 2GB",
        0x0f: "Microsoft Extended LBA",
    }   # this dict could be extended to include more types

    def __init__(self, data=bytes(16), *, 
                 bootable=None, chs_start=None, type=None, chs_end=None,
                 start=None, size=None
        ):
        """
        :param data: raw bytes for the MBR record, at least 16 bytes long
        """
        def _chs(value):
            """A little helper for CHS parsing."""
            head, sector, cylinder = struct.unpack("<3B", value)
            cylinder = ((sector & 0b11000000) << 2) | cylinder
            sector   =   sector & 0b00111111
            return (cylinder, head, sector)

        p_bootable  = data[0]
        p_chs_start = _chs(data[1:4])
        p_type      = data[4]
        p_chs_end   = _chs(data[5:8])
        p_start,    = struct.unpack("<I", data[8:12])
        p_size,     = struct.unpack("<I", data[12:16])
        # and now we set, either the overridden values or the parsed ones
        self._raw_data = data
        self.bootable  = bootable or p_bootable
        self.chs_start = chs_start or p_chs_start
        self.type      = type or p_type
        self.chs_end   = chs_end or p_chs_end
        self.start     = start or p_start
        self.size      = size or p_size
    
    def __bytes__(self):
        """
        Handles packing the partitions attributes into the binary struct that
        is saved to disk.
        """
        def _chs(cylinder, head, sector):
            sector   = ((cylinder & 0b1100000000) >> 2) | (sector & 0b00111111)
            cylinder =   cylinder & 0b0011111111
            # just in case, let's clip the values
            cylinder = min(max(0, cylinder), 255)
            head     = min(max(0, head), 255)
            sector   = min(max(0, sector), 255)
            return (cylinder, head, sector)
        
        start_c, start_h, start_s = _chs(*self._chs_start)
        end_c  , end_h,   end_s   = _chs(*self._chs_end)
        return struct.pack(
            "<BBBBBBBBII",
            self._bootable,
            start_h, start_s, start_c,
            self._type,
            end_h, end_s, end_c,
            self._start,
            self._size
        )
    
    def __repr__(self):
        ret = f"< Partition - {self.type} - boot: {self.bootable} @ {self.start} of {self.size} >"
        return ret
    
    @property
    def bootable(self):
        """
        Bootable flag for the partition.

        When setting this property, the boolean value (or expression) passed is
        used to determine the correct bytes to store on disk.
        """
        return self._bootable == 0x80
    
    @bootable.setter
    def bootable(self, value):
        if value:
            self._bootable = 0x80
        else:
            self._bootable = 0x0
    
    @property
    def chs_start(self):
        """
        Cylinder-Head-Sector start address. Parsed at initialization time.

        Read only.
        """
        return self._chs_start
    
    @chs_start.setter
    def chs_start(self, value):
        self._chs_start = value
    
    @property
    def type(self):
        """
        Partition type. The getter returns a string coded in the
        Partition.partition_types dict. Partition._type holds the underlying int
        value.

        The setter clips any integer value given to the 0..255 range.
        """
        return self.partition_types.get(self._type, "Unknown")
    
    @type.setter
    def type(self, value):
        value = int(value)
        value = min(max(0, value), 255)  # clip it to 0..255
        self._type = value
    
    @property
    def chs_end(self):
        """
        Cylinder-Head-Sector end address. Parsed at initialization time.

        Read only.
        """
        return self._chs_end
    
    @chs_end.setter
    def chs_end(self, value):
        self._chs_end = value

    @property
    def start(self):
        """
        LBA start address. Parsed at initialization time. The setter enforces a
        0x0..0xffffffff range.
        """
        return self._start
    
    @start.setter
    def start(self, value):
        value = int(value)
        value = min(max(0, value), 0xffffffff)
        self._start = value
    
    @property
    def size(self):
        """
        Parition size. Parsed at initialization time. The setter enforces a
        0x0..0xffffffff range.
        """
        return self._size
    
    @size.setter
    def size(self, value):
        value = int(value)
        value = min(max(0, value), 0xffffffff)
        self._size = value
    
    def to_bytes(self):
        return bytes(self)


class Table:
    """
    Handles one Master Boot Record (simple case without Extended Partitions).
    """
    def __init__(self, data=bytes(512), *,
                 boot_code=None, partitions=None
        ):
        """
        :param data: raw bytes read from an MBR (at least 512 bytes long)
        """
        self._raw_data = data
        data = data[446:]
        p_boot_code = self._raw_data[:446]
        p_parts = filter(
            lambda x: x != b"\x00" * 16,  # filter out empty partitions
            (data[i * 16: i * 16 + 16] for i in range(4))
        )
        p_partitions = [Partition(p) for p in p_parts]
        # this implementation is "good enough", but not thorough
        # -- the legacy files have a larger implementation that includes
        #    extended partitions parsing
        # override parsed values with keyword-only arguments
        self.boot_code = boot_code or p_boot_code
        self.partitions = partitions or p_partitions

    def __bytes__(self):
        ret = bytearray(512)
        ret[:446] = self.boot_code
        parts = b"".join(bytes(p) for p in self.partitions)
        parts = parts[:64]
        psize = len(parts)
        ret[446: 446 + psize] = parts
        ret[510: 512] = b"\x55\xaa"
        return bytes(ret)
    
    def __repr__(self):
        ret  = [f"< MBR Partition Table @ {id(self)} >"]
        ret.extend(
               [f"    {p}" for p in self.partitions]
        )
        return "\n".join(ret)

    def to_bytes(self):
        return bytes(self)



"""
Basic VHD format support to test making a full, fixed size, working VHD file
from scratch.

References:
https://github.com/libyal/libvhdi/blob/master/documentation/Virtual%20Hard%20Disk%20(VHD)%20image%20format.asciidoc
https://www.forensicfocus.com/articles/virtual-hard-disk-image-format-a-forensic-overview/
"""
import datetime
import struct


from .guid import GUID


class VHDFooter:
    """
    VHD Fixed Disk Footer to be found at the end of a VHD file (raw image + 
    VHD Header).
    """

    vhd_disk_types = {
        0: "None",
        1: "Reserved (deprecated) #1",
        2: "Fixed hard disk",
        3: "Dynamic hard disk",
        4: "Differential hard disk",
        5: "Reserved (deprecated) #5",
        6: "Reserved (deprecated) #6",
    }

    def __init__(self, data=bytes(512), *, 
                 cookie=None, features=None, format_version=None, next_offset=None,
                 modification_time=None, creator_app=None, creator_version=None,
                 creator_host=None, disk_size=None, data_size=None,
                 disk_geometry=None, disk_type=None, checksum=None,
                 identifier=None, saved_state=None
        ):
        """
        :param data: the 512 bytes that make the VHD Fixed Footer
        """
        self._raw_data = data
        # only the first 85 bytes have information, the rest is/should be \x00
        cook, feat, form, nxto = struct.unpack(">8sLLq",   data[  : 24])
        mtim, capp, cver, chos = struct.unpack(">L4sL4s",  data[24: 40])
        disz, dasz, dgeo, dtyp = struct.unpack(">QQ4sL",   data[40: 64])
        csum, guid, stat       = struct.unpack(">l16sB",   data[64: 85])
        # and now a few transformations before we set it all into the attributes
        mtim = datetime.datetime(2000, 1, 1) + datetime.timedelta(0, mtim)
        cver_major = cver >> 16
        cver_minor = cver & 0x0000ffff
        cver = (cver_major, cver_minor)
        cyl, head, sec = struct.unpack(">HBB", dgeo)
        dgeo = (cyl, head, sec)
        # and now we set the _ attributes for the properties
        self.cookie            = cookie or cook 
        self.features          = features or feat
        self.format_version    = format_version or form
        self.next_offset       = next_offset or nxto
        self.modification_time = modification_time or mtim
        self.creator_app       = creator_app or capp
        self.creator_version   = creator_version or cver
        self.creator_host      = creator_host or chos
        self.disk_size         = disk_size or disz
        self.data_size         = data_size or dasz
        self.disk_geometry     = disk_geometry or dgeo
        self.disk_type         = disk_type or dtyp
        self.checksum          = csum  # we're not overriding the checksum
        self.identifier        = identifier or GUID(guid)
        self.saved_state       = saved_state or stat
    
    def __bytes__(self):
        data = bytearray()
        # a few of the properties need some preprocessing before being packed
        # as bytes
        modtime = self.modification_time - datetime.datetime(2000, 1, 1)
        modtime = (modtime.days * 86400) + modtime.seconds
        cver_major, cver_minor = self.creator_version
        cver = (cver_major << 16) | cver_minor
        cyl, head, sec = self.disk_geometry
        dgeo = struct.pack(">HBB", cyl, head, sec)
        # ok, now we're ready to pack!
        data += struct.pack(
            ">8sLLqL4sL4sQQ4sLl16sB",
            self.cookie,
            self.features,
            self.format_version,
            self.next_offset,  
            modtime,
            self.creator_app,
            cver,
            self.creator_host, 
            self.disk_size,    
            self.data_size,       
            dgeo,
            self._disk_type,   
            0,                     # checksum, we calculate it on the spot
            bytes(self.identifier),
            self.saved_state
        )
        checksum = ~sum(data)
        data[64:68] = struct.pack(">l", checksum)
        data += bytes(427)
        return bytes(data)

    @property
    def cookie(self):
        """Cookie. Read only, fixed as b"conectix"."""
        return self._cookie
    
    @cookie.setter
    def cookie(self, value):
        value = bytes(value)[:8]
        if len(value) != 8:
            value += bytes(8 - len(value))
        self._cookie = value

    @property
    def features(self):
        """Features. Read only, fixed as 2."""
        return self._features
    
    @features.setter
    def features(self, value):
        value = int(value)
        value = max(min(value, 0xffffffff), 0)
        self._features = value

    @property
    def format_version(self):
        """Format version. Should always be 0x00010000."""
        return self._format_version
    
    @format_version.setter
    def format_version(self, value):
        value = int(value)
        value = max(min(value, 0xffffffff), 0)
        self._format_version = value

    @property
    def next_offset(self):
        """Next offset. Should always be as -1."""
        return self._next_offset
    
    @next_offset.setter
    def next_offset(self, value):
        self._next_offset = value

    @property
    def modification_time(self):
        """
        Modification time. Second resolution. Can't store datetimes before 
        2000-01-01.
        """
        return self._modification_time
    
    @modification_time.setter
    def modification_time(self, value):
        value = datetime.datetime.replace(value, microsecond=0)
        if value < datetime.datetime(2000, 1, 1):
            value = datetime.datetime(2000, 1, 1)
        self._modification_time = value

    @property
    def creator_app(self):
        """
        Creator application. Gets encoded as bytes with "ascii" encoding, and
        capped at 4 bytes long.
        """
        return self._creator_app
    
    @creator_app.setter
    def creator_app(self, value):
        if isinstance(value, str):
            value = value.encode("ascii")
        if len(value) < 4:
            value += b"\x20" * (4 - len(value))  # extend with spaces
        self._creator_app = value
    
    @property
    def creator_version(self):
        """
        Creator version. Tuple of ints, get encoded as 2 16-bit values glued
        together into a 32-bit int.
        """
        return self._creator_version
    
    @creator_version.setter
    def creator_version(self, value):
        major, minor = value
        major, minor = int(major), int(minor)  # just making sure
        if major > 0xffff:
            major = 0xffff
        if minor > 0xffff:
            minor = 0xffff
        self._creator_version = (major, minor)
    
    @property
    def creator_host(self):
        """
        Creator host. Gets encoded as bytes with "ascii" encoding, and capped at
        4 bytes long.
        """
        return self._creator_host
    
    @creator_host.setter
    def creator_host(self, value):
        if isinstance(value, str):
            value = value.encode("ascii")
        if len(value) < 4:
            value += b"\x20" * (4 - len(value))  # extend with spaces
        value = value[0:4]
        self._creator_host = value

    @property
    def disk_size(self):
        """Disk size. 64-bit int."""
        return self._disk_size
    
    @disk_size.setter
    def disk_size(self, value):
        value = int(value)  #   just making sure...
        value = max(value, 0)
        value = min(value, 0xffffffffffffffff)
        self._disk_size = value
    
    @property
    def data_size(self):
        """Data size. 64-bit int."""
        return self._data_size
    
    @data_size.setter
    def data_size(self, value):
        value = int(value)  #   just making sure...
        value = max(value, 0)
        value = min(value, 0xffffffffffffffff)  # and a bit of clipping
        self._data_size = value
    
    @property
    def disk_geometry(self):
        """Disk geometry in as a (cylinder, head, sector) tuple of ints."""
        return self._disk_geometry
    
    @disk_geometry.setter
    def disk_geometry(self, value):
        cylinder, head, sector = map(int, value)
        cylinder = min(max(cylinder, 0), 65535)
        head     = min(max(head,   0), 255)
        sector   = min(max(sector, 0), 255)
        # maybe throw an error on these cases? shouldn't happen, but still
        self._disk_geometry = (cylinder, head, sector)
    
    @property
    def disk_type(self):
        """
        Disk type. See VHDFooter.vhd_disk_types for valid values.
        """
        k = self._disk_type
        v = self.vhd_disk_types.get(k, "Unknown")
        return v
    
    @disk_type.setter
    def disk_type(self, value):
        value = int(value)
        if not(0 <= value <= 6):
            raise ValueError("the given value is not in the 0..6 range")
        self._disk_type = value
    
    @property
    def checksum(self):
        """
        Checksum. Gets recalculated on conversion of footer to bytes.
        """
        return self._checksum

    @checksum.setter
    def checksum(self, value):
        self._checksum = value

    @property
    def identifier(self):
        """Identifier GUID."""
        return self._identifier
    
    @identifier.setter
    def identifier(self, value):
        if not(isinstance(value, GUID)):
            raise ValueError(f"expected GUID, got {type(identifier)} instead")
        self._identifier = value
    
    @property
    def saved_state(self):
        """
        Saved state. Flag to indicate the image is in saved staty. 8-bit int.
        """
        return self._saved_state
    
    @saved_state.setter
    def saved_state(self, value):
        value = int(value)
        value = min(max(value, 0), 255)
        self._saved_state = value
    
    def __repr__(self):
        r = (f"< VHDFooter: {self.disk_type}"
             f"{self.disk_size} (last modified {self.modification_time})>"
            )
        return r
    
    def __str__(self):
        ret = [
            f"< VHDFooter @ {id(self)} >",
            f"           cookie: {str(self.cookie):>20}",
            f"         features: {str(self.features):>20}",
            f"   format version: {str(self.format_version):>20}",
            f"      next offset: {str(self.next_offset):>20}",
            f"modification time: {str(self.modification_time):>20}",
            f"      creator app: {str(self.creator_app):>20}",
            f"  creator version: {str(self.creator_version):>20}",
            f"     creator host: {str(self.creator_host):>20}",
            f"        disk size: {str(self.disk_size):>20}",
            f"        data size: {str(self.data_size):>20}",
            f"    disk geometry: {str(self.disk_geometry):>20}",
            f"        disk type: {str(self.disk_type):>20}",
            f"         checksum: {str(self.checksum):>20}",
            f"       identifier: {str(self.identifier):>20}",
            f"      saved state: {str(self.saved_state):>20}",
        ]
        return "\n".join(ret)



"""
Module to parse GUID Partition Tables.
"""
import struct


from .guid import GUID
from .utils import slicer


class Partition:
    """GPT Partition Record"""
    partition_types = {
        # we have to fill in a few GUIDs here...
        # the only sane list we've found already is here:
        # * http://www.liquisearch.com/guid_partition_table/partition_type_guids
        GUID("C12A7328-F81F-11D2-BA4B-00A0C93EC93B"): "EFI System Partition",
        # Microsoft GUID types
        GUID("DE94BBA4-06D1-4D40-A16A-BFD50179D6AC"): "Windows Recovery Tools Partition",
        GUID("E3C9E316-0B5C-4DB8-817D-F92DF00215AE"): "Microsoft Reserved Partition",
        GUID("EBD0A0A2-B9E5-4433-87C0-68B6B72699C7"): "Basic Data Partition",
    }

    def __init__(self, data=bytes(128), length=128, *,
                 guid_type=None, guid_part=None, start=None, end=None,
                 size=None, flags=None, name=None
        ):
        """
        :param data: raw bytes of the partition record
        :param length: length in bytes of the partition record, defaults to 128
        """
        self._raw_data = data
        self._length   = length
        # now for something extra (and not completely different)
        data   = self._raw_data
        nlength = self._length - 56  # this is the length of the name field
        p_guid_type = GUID(data[ 0:16], mixed_endian=True)
        p_guid_part = GUID(data[16:32], mixed_endian=True)
        p_start, p_end, p_flags = struct.unpack("<QQQ", data[32:56])
        p_name, = struct.unpack(f"{nlength}s", data[56:])
        p_name  = p_name.decode("utf-16")
        if "\x00" in p_name:
            p_name = p_name[:p_name.find("\x00")]
        # and now we set the values, parsed or given
        self.guid_type = guid_type or p_guid_type
        self.guid_part = guid_part or p_guid_part
        self.type      = self.partition_types.get(self.guid_type, "Unknown")  # make property
        self.start     = start or p_start
        self.end       = end or p_end
        self.size      = self.end - self.start + 1
        self.name      = name or p_name
    
    def __repr__(self):
        ret = "< Partition - {type} - @ {start} of {size} >"
        return ret.format_map(self.__dict__)


class Table:
    """GPT Partition Table"""
    def __init__(self, raw_data, *,
                 signature=None, revision=None, header_size=None,
                 current_lba=None, backup_lba=None, first_lba=None,
                 last_lba=None, disk_guid=None, parts_lba=None, parts_num=None,
                 parts_size=None, crc32_parts=None, partitions=None
        ):
        """

        """
        self._raw_data = raw_data
        p_rev, p_hsize, p_crc32, _res, p_cur_lba, p_back_lba, p_first_lba, p_last_lba \
            = struct.unpack("<LLLLQQQQ", raw_data[8:56])
        # disk GUID comes in the middle here...
        p_parts_lba, p_parts_num, p_parts_size, p_crc32 = struct.unpack("<QLLL", raw_data[72:92])
        # and now we set them all...
        self.signature   = signature or raw_data[0:8]
        self.revision    = revision or p_rev
        self.header_size = header_size or p_hsize
        self.current_lba = current_lba or p_cur_lba
        self.backup_lba  = backup_lba or p_back_lba
        self.first_lba   = first_lba or p_first_lba
        self.last_lba    = last_lba or p_last_lba
        self.disk_guid   = disk_guid or GUID(raw_data[56:72], mixed_endian=True)
        self.parts_lba   = parts_lba or p_parts_lba
        self.parts_num   = parts_num or p_parts_num
        self.parts_size  = parts_size or p_parts_size
        self.crc32_parts = crc32_parts or p_crc32
        part_data = raw_data[512: 512 + 16384]  # should be this length...
        p_partitions = [
            Partition(s, self.parts_size)
            for s in slicer(part_data, self.parts_size)
            if s != bytes(self.parts_size)
        ]
        self.partitions  = partitions or p_partitions
    
    def __repr__(self):
        ret  = [f"< GUID Partition Table @ {id(self)} >"]
        ret.extend(
               [f"    {p}" for p in self.partitions]
        )
        return "\n".join(ret)
    
    def __str__(self):
        ret = [
            f"< GPT Partition Table @ {id(self)} >",
            f"    signature: {str(self.signature):>20}",
            f"     revision: {str(self.revision):>20}",
            f"  header size: {str(self.header_size):>20}",
            f"  current LBA: {str(self.current_lba):>20}",
            f"   backup LBA: {str(self.backup_lba):>20}",
            f"    first LBA: {str(self.first_lba):>20}",
            f"     last LBA: {str(self.last_lba):>20}",
            f"    disk GUID: {str(self.disk_guid):>20}",
            f"    parts LBA: {str(self.parts_lba):>20}",
            f"      parts #: {str(self.parts_num):>20}",
            f"   parts size: {str(self.parts_size):>20}",
        ]
        return "\n".join(ret)



"""
General purpose functions that are used in different places."
"""
import string

def hexshow(data):
    """
    Pretty prints in the console the bytes object it takes as parameter.

    :param data: a bytes sequence to pretty print on the screent
    """
    lines = [data[i * 16: i * 16 + 16] for i in range(len(data) //16)]
    for i, line in enumerate(lines):
        out = f"{i * 16:08x} | "
        out += " ".join( f"{b:02x}" for b in line )
        out += " "
        out += "".join(
            f"{i}"
            for i in map(
                lambda x: chr(x) if (chr(x) in string.printable and chr(x) not in "\n\r") else ".", line))
        print(out)

def slicer(iterable, length):
    """Slices an iterable into object into length sized chunks."""
    full_length = len(iterable)
    part_length = (full_length // length)
    for i in range(part_length):
        yield iterable[i * length: i * length + length]


def str2bytes(data):
    """
    Converts a str with a bytes sequence into said sequence. If the given
    sequence is not even-sized, the last character is ignored.

    :param data: string with a byte-sequence in it (example: "ffd8")
    :returns: bytes object of that sequence (example: b"\\xff\\xd8")
    """
    return bytes.fromhex(data)



"""
This module handles reading from an Ext2 filesystem.
"""

# For the future: implement the writing capability, and the parser of the ext3's journal.

# To check:
#  - Check the path that I send to Ext2.open(), because I think it is assumed that
#    we must put the 'name' of the partition (which would represent the root directory '/')
#    followed by what we want to open (but we can even put anything, or directly nothing;
#    I have to test that well). In Windows it would be for example C:\ ..., being C: the name/label
#    of the volume/partition (or the mount point), it would be necessary to see what format it is in Linux (it could be sda1/...).
#  - I could implement the methods that I left defined at the end of the 'superblock.py' and 'inode.py' files.

# NOTE: Whenever we speak about a "block number" or "inode number", it is spoken in an absolute way, that is,
#       that block or inode number in the entire filesystem (and if it is relative to something, it will be specified).
#       The blocks are counted from the boot area to the end of the partition, and the inodes are counted for each entry
#       in the inode table of each block group.

from . import directory_entry
from . import group_descriptor
from . import inode
from . import superblock

import struct
from math import ceil

# Constants representing sizes in bytes
DISK_SECTOR_SIZE        = 512

SB_STRUCT_SIZE          = 1024
GD_STRUCT_SIZE          = 32
# INODE_STRUCT_SIZE       = 128
#   this value actually depends on the revision of ext2
#       rev. 0: always 128 bytes
#       rev. 1: must be a perfect power of 2 and must be smaller or equal to the block size
#   anyway, at the moment I only parse the first 128 bytes of the inode (in 'Inode.py'), which is the base size of any inode
#   but I need the exact value to correctly read the inodes from the disk (I get this value from the superblock, see Ext2's constructor).
DENTRY_STRUCT_BASE_SIZE = 8

# Encoding of filenames and paths
ENCODING = 'latin-1'

class Directory:
    """
    A directory is actually an inode whose data blocks that it points to,
    contain directory entries.
    What I do with this class is get to those blocks and parse them.

    :param filesystem: the base filesystem (of class Ext2) to which the Directory belongs
    :param inode_obj: the inode (Inode object) that represents the directory
    :param name: the directory name (it does not come from its inode, but from some directory entry of its parent directory)
    :param parent: the parent (Directory object) of the (new) directory, used to build a path string
    """
    def __init__(self, filesystem, inode_obj, name, parent=None):
        
        self.filesystem = filesystem
        self.inode_obj = inode_obj
        self.name = name

        self.files = []

        if parent is None:
            self.path = self.name.decode(ENCODING) # for the root, 'name' will be the 'volume_name'.
        else:
            self.path = parent.path + '/' + self.name.decode(ENCODING)

        self._parse()

    def __repr__(self):
        # a bit of a hack (X or Y) in case we want to show the root directory,
        # and it doesn't have a certain name (volume_name == "" (empty))
        return f"< Directory {self.path or '/'} >"

    def __str__(self):
        ret = f"< Directory {self.path or '/'} >\n"
        ret += "---------------------------------FILES------------------------------------\n"
        for f in self.files:
            ret += f"{f.name.decode(ENCODING):50} ({f.file_type})\n"
        ret += "--------------------------------------------------------------------------\n"
        return ret

    def _read_dentries(self, block_number):
        """
        Internal method that reads and parses the directory entries of a block
        in the filesystem. It adds DirectoryEntry objects to the list of files
        of the current Directory object.
        """

        raw_block = self.filesystem.read_block(block_number)

        offset = 0 # this variable will store the starting address of each directory entry
        block_size = self.filesystem.superblock.s_log_block_size
        # let's read and parse the directory entries (the last valid entry points to the end of the block)
        while offset < block_size:
            dx = offset + DENTRY_STRUCT_BASE_SIZE
            file = directory_entry.DirectoryEntry(raw_block[offset:dx]) # we parse the first 8 bytes (the constant part of the dentry).
            file.name = raw_block[dx:dx+file.name_len] # and now the rest, corresponding to the name (the variable part).
            if file.inode != 0:
                # This happens only if the first file is deleted;
                # all other deleted file entries will be skipped due to
                # a proper rec_len of the previous entry.
                # (file deleted -> prev. rec_len points to next dentry, and inode=0)
                self.files.append(file)
            offset += file.rec_len

    def _parse(self):
        """
        Method that parses the pointers of the direct and indirect blocks pointed by
        the inode that represents a directory, in order to obtain the number of each
        data block assigned to the directory and parse the directory entries they contain.
        """

        # let's read the direct blocks
        direct_blocks = self.filesystem._parse_direct_blocks(self.inode_obj.i_block[0:12])
        for directory_block_number in direct_blocks:
            self._read_dentries(directory_block_number)

        # and now the indirect blocks
        if self.inode_obj.i_block[12] != 0: # thus, we avoid reading the block '0' (*)
            indirect_1_block = self.filesystem._parse_indirect_1_block(self.inode_obj.i_block[12])
            for directory_block_number in indirect_1_block:
                self._read_dentries(directory_block_number) # we read the directory entries contained in the blocks pointed to by the pointers of indirect_1_block.

        if self.inode_obj.i_block[13] != 0:
            indirect_2_block = self.filesystem._parse_indirect_2_block(self.inode_obj.i_block[13])
            for directory_block_number in indirect_2_block:
                self._read_dentries(directory_block_number)

        if self.inode_obj.i_block[14] != 0:
            indirect_3_block = self.filesystem._parse_indirect_3_block(self.inode_obj.i_block[14])
            for directory_block_number in indirect_3_block:
                self._read_dentries(directory_block_number)


        # *: Since a pointer pointing to block 0, means null pointer.
        #    As the filesystem allocates new blocks to the directory/file, the pointers will no longer be null.
        #    If we read block 0, we would be reading the boot area or boot area + superblock, depending on the block_size.
        #
        #    And when we analyze the indirect blocks, there will be no problem of 0's
        #    because we will filter them. In addition, that an indirect pointer of
        #    the inode is !=0, it means that there will be minimally a sequence of
        #    non-null pointers and there will be no problem of 0's.
        #    (for example, if i_block[14] != 0, it means that there is at least
        #    one data block assigned to the file at the end of
        #    triple_block -> double_block -> simple_block -> data_block)

        # NOTE: Although I store in variables the list of the numbers of data blocks
        #       that contain the directory entries, they are local to this internal method,
        #       so when finished, they are 'freed' of memory, those references are not preserved
        #       (I say this due to the magnitude of bytes that we could have, since each block is referenced by a 4-byte pointer).

    def show_dentries(self):
        """
        Method that returns a string with the representation of each
        DirectoryEntry object that contains the list of files in the directory.
        """
        ret = f"< Directory {self.path or '/'} >\n"
        ret += "----------------------------DIRECTORY-ENTRIES-----------------------------\n"
        for f in self.files:
            ret += f"{str(f)}\n"
        ret += "--------------------------------------------------------------------------\n"
        return ret

    def show_inode(self):
        """
        Method that returns a string with the representation of the Inode object
        associated with the directory.
        """
        ret = f"< Directory {self.path or '/'} >\n"
        ret += "-----------------------------ASSOCIATED-INODE-----------------------------\n"
        ret += f"{str(self.inode_obj)}"
        ret += "--------------------------------------------------------------------------\n"
        return ret


class FileHandle:
    """
    Class that handles a data file, given its inode that represents it.
    It will have the basic file operations: read, write, seek, close ...
    (the 'open' would be implemented in the Ext2 class, since it just returns
    this FileHandle object, as when we make a true open in Python or another language).
    At the moment, the 'write' will not be implemented.

    :param filesystem: the base filesystem (of class Ext2) to which the file belongs
    :param inode_obj: the inode (Inode object) that represents the file
    :param name: the file name (it does not come from its inode, but from some directory entry of its parent directory)
    :param parent: the directory (Directory object) that "contains" the file, used to build a path string
    """
    def __init__(self, filesystem, inode_obj, name, parent=None):
        
        self.filesystem = filesystem
        self.inode_obj = inode_obj
        self.name = name

        if parent is None:
            self.path = self.name.decode(ENCODING) # for the root, 'name' will be the 'volume_name'.
        else:
            self.path = parent.path + '/' + self.name.decode(ENCODING)

        self.closed = False # I should put this attribute as 'private' through properties.

        # We will always keep the file's bytes in a 'buffer',
        # one block at a time (starting with its first block)
        self._buffer      = self.filesystem.read_block(self.inode_obj.i_block[0])
        self._buffer_pos  = 0 # 0 <= pos < buffer_size
        self._buffer_size = self.filesystem.superblock.s_log_block_size

        # We keep the pointer to the last 'buffered' block (block number relative to the file),
        # so that in the next reading of the file, we know where we are.
        # (remember that we will implement a 'true' read, that is, we can read
        # a file on different occasions and the byte pointer will move)
        self._current_block_pointer = 0

        # position in the file (pointer to the byte number), .tell() returns this
        self._file_pos = 0 # 0 <= pos < self.inode_obj.i_size

        # We obtain in advance all the block numbers that are assigned to the file
        # (I don't think it is the most efficient in terms of memory, but it makes the 'read' method much easier)
        # (remember that each block is referenced by a 4-byte pointer, so 4*len(list) could be very large)
        # (here we can see different maximum sizes in blocks of a file: https://www.nongnu.org/ext2-doc/ext2.html#def-blocks)
        self._data_block_numbers = []
        self._data_block_numbers = self.filesystem._parse_direct_blocks(self.inode_obj.i_block[0:12])
        if self.inode_obj.i_block[12] != 0: # thus, we avoid reading the block '0' (just like in the class Directory)
            self._data_block_numbers += self.filesystem._parse_indirect_1_block(self.inode_obj.i_block[12])
        if self.inode_obj.i_block[13] != 0:
            self._data_block_numbers += self.filesystem._parse_indirect_2_block(self.inode_obj.i_block[13])
        if self.inode_obj.i_block[14] != 0:
            self._data_block_numbers += self.filesystem._parse_indirect_3_block(self.inode_obj.i_block[14])
        # the 'end of file' is defined by the length of this list (the last position would be its last data block).

    def __repr__(self):
        return f"< FileHandle for {self.path} >"
    
    def __str__(self):
        return self.__repr__()

    def close(self):
        """
        Closes the file.
        """
        if self.closed == True:
            raise ValueError("I/O operation on closed file.")

        self.closed = True

        return True

    def read(self, size=-1):
        """
        Method that reads a certain number of bytes (size) from the file
        referenced by the FileHandle object (as long as the file is 'open').
        If the 'size' argument is omitted, the entire file will be read
        (or the remaining bytes of it depending on the position of its pointer).
        """
        if self.closed == True:
            raise ValueError("I/O operation on closed file.")

        if size < 0: # it would indicate that we want to read the remaining bytes of the file.
            size = self.inode_obj.i_size - self._file_pos # file length in bytes - current position of the file pointer.

        ret = [] # here we will store the byte strings that we read from the buffer, and in the end we will join everything.

        # while it is necessary to read a next block of the file to satisfy the reading, and we are not already in its last block (end of file)...
        while size > (self._buffer_size - self._buffer_pos) and self._current_block_pointer+1 < len(self._data_block_numbers): # len is O(1)
            ret.append(self._buffer[self._buffer_pos:]) # here we start from 'buffer_pos' in case it was read a little first (and we did not enter this 'while'), and then a lot (and we did enter).
            size           -= self._buffer_size - self._buffer_pos # we are subtracting from 'size' the number of bytes that we already read.
            self._file_pos += self._buffer_size - self._buffer_pos # we advance the file pointer as we read.
            self._buffer_pos = 0 # as we consume the entire buffer, we set its position to 0.
            next_block_pointer = self._current_block_pointer + 1
            next_block_number = self._data_block_numbers[next_block_pointer]
            self._buffer = self.filesystem.read_block(next_block_number) # we read the next data block and load it into the buffer.
            self._current_block_pointer = next_block_pointer

        span = min(self.inode_obj.i_size - self._file_pos, size) # We do this to avoid reading more bytes from the buffer than the remaining bytes of the file.
        ret.append(self._buffer[self._buffer_pos:self._buffer_pos+span]) # we read the X ('span') bytes from the buffer
        self._buffer_pos += span # and we move the pointers of both the buffer and the file.
        self._file_pos   += span # (there will be no overflow problems because 'span' will always be smaller than the block size)
        return b"".join(ret)

            # I have to check what to do if trying to read more bytes than the file has. DONE
            # (there will come a time when we reach the last block, the while will be cut there,
            # and the 'span' will be left with the right amount of bytes that remain to be read)
            # (as in Python, it reads up to where there are bytes and that's it. And if we continue to want to read later, it returns b'' [empty])
            
            # And also what would happen if I read a few bytes at a time, avoiding entering to the while,
            # and there comes a time that I go over of the block size; since I would need to read the following block. DONE
            # (this is solved by checking in the while '(self._buffer_size - self._buffer_pos)',
            # thanks to the position of the pointer we will know when it will be necessary to go to the next block,
            # and in those cases we enter to the while, since 'size' will be greater than that difference)

            # I would have to check what happens if I have already read the entire file, and I want to read such a large number of bytes
            # that it goes into the while, because there, an extra append would be made with what was left in the buffer. DONE
            # (It would not enter the while because the second condition would give 'false'; we will have stayed in the position of its last block).

    def seek(self, offset, whence=0):
        """
        Moves the file pointer, based on 'offset'.
        The 'whence' argument is optional and defaults to 0, which means absolute file positioning,
        other values are 1 which means seek relative to the current position,
        and 2 means seek relative to the file's end.
        Returns the current position of the file pointer, after moving.
        """
        
        if self.closed == True:
            raise ValueError("I/O operation on closed file.")

        if whence == 0:
            new_file_pos = offset
        elif whence == 1:
            new_file_pos = self._file_pos + offset
        elif whence == 2:
            new_file_pos = self.inode_obj.i_size + offset # the 'offset' is expected to be negative
        else:
            raise ValueError(f"invalid whence ({whence}, should be 0, 1 or 2)")

        # The idea is to calculate the number of blocks that the new ABSOLUTE position
        # of the file pointer represents (always starting the count from the beginning of the file).
        # As we have previously loaded each data block number of the file (self._data_block_numbers),
        # we can directly obtain the block where the new position falls, and thus load it into the buffer.

        if new_file_pos >= self.inode_obj.i_size:       # If we want to go beyond the end of the file
            block_number = self._data_block_numbers[-1] # we will keep its last block
            self._buffer_pos = self._buffer_size        # and move the buffer pointer to the end of the buffer
            self._file_pos = self.inode_obj.i_size      # and the file pointer at the end of the file.
            self._current_block_pointer = len(self._data_block_numbers) - 1 # We update this attribute for a future 'read()'.
        elif new_file_pos < 0:
            block_number = self._data_block_numbers[0] # A similar analysis if we want to go further back of the beginning of the file..
            self._buffer_pos = 0
            self._file_pos = 0
            self._current_block_pointer = 0
        else:
            block_index = new_file_pos // self._buffer_size       # Else, we get how many blocks of the file the offset covers,
            block_number = self._data_block_numbers[block_index]  # we keep the data block number obtained from the calculated quantity
            self._buffer_pos = new_file_pos % self._buffer_size   # we move the buffer pointer the remaining amount of bytes
            self._file_pos = new_file_pos                         # and the file pointer should be the offset.
            self._current_block_pointer = block_index             # We update this attribute for a future 'read()'.
        
        self._buffer = self.filesystem.read_block(block_number) # Finally we update the buffer with the reading of the obtained block.
        
        return self._file_pos

        # NOTA: I assume that in the list of data_block_numbers there would be no 'gaps' (0's) in between.
        #       I don't know if it's possible *, but if it happens, this would fail
        #       since data_block_numbers[block_index] would be out of date.
        #       (* that is, I don't know if it is possible for a data block, to be deallocated from the file,
        #       leaving a pointer to null (->0) in between, and then the inode pointers are not 'defragmented', or not 'compact')

    def tell(self):
        """
        Returns the current position of the file pointer.
        """
        if self.closed == True:
            raise ValueError("I/O operation on closed file.")

        return self._file_pos

    def show_inode(self):
        """
        Method that returns a string with the representation of the Inode object
        associated with the file.
        """
        ret = f"< File {self.path} >\n"
        ret += "-----------------------------ASSOCIATED-INODE-----------------------------\n"
        ret += f"{str(self.inode_obj)}"
        ret += "--------------------------------------------------------------------------\n"
        return ret


class Ext2:
    """
    Class that handles an ext2 filesystem.

    :param path: It is the way to access the storage device.
        On Windows, it is in the format: r"\\.\PhysicalDriveX", and on Linux "/dev/sdX".
    :param base_address: It is the offset (byte number), inside the device, where
        the partition (with an ext2 fs) that we want to analyze, begins.
    """
    def __init__(self, path, base_address):
        """
        The storage device is opened.
        The byte pointer is positioned at the beginning of the desired partition.
        And the boot area, the superblock (original), the group descriptor table
        (original) and the root inode, are read.
        """
        self.path = path
        self.base_address = base_address

        self.handle = open(path, "rb") # read-only for now (for security, I should set this attribute as private through properties)
        self.handle.seek(self.base_address)

        # the first 2 sectors of the partition correspond to the boot area (are unused by the ext2 filesystem).
        self.boot_area = self.handle.read(DISK_SECTOR_SIZE*2)
        # the next 1024 bytes correspond to the original superblock (we are already within block group 0).
        self.superblock = superblock.Superblock(self.handle.read(SB_STRUCT_SIZE))
        # and then there will be as many group descriptors as there are block groups in the filesystem.
        # (the group descriptor table begins at the block following the superblock)
        self.handle.seek(self.base_address + self.superblock.s_log_block_size * (self.superblock.s_first_data_block + 1))
        useful_blocks = self.superblock.s_blocks_count - self.superblock.s_first_data_block # if block_size=1K, the first block doesn't belong to the first block_group
        block_group_count = ceil(useful_blocks / self.superblock.s_blocks_per_group) # (*1)
        self.group_descriptors = [group_descriptor.GroupDescriptor(self.handle.read(GD_STRUCT_SIZE)) for i in range(block_group_count)] # (*2)

        # we get the inode size, since it is not always 128 bytes.
        self.INODE_STRUCT_SIZE = self.superblock.s_inode_size

        # The root directory always corresponds to inode No. 2 (inode_table[1]), which belongs to block group No. 1 (bg[0]).
        # (note: I don't use the 'read_inode' method cause for me in this case it is more descriptive to see the read_record's arguments)
        inode_table_block = self.group_descriptors[0].bg_inode_table
        raw_root_inode_entry = self.read_record(inode_table_block, self.INODE_STRUCT_SIZE, offset=self.INODE_STRUCT_SIZE)
        root_inode = inode.Inode(raw_root_inode_entry)
        self.root = Directory(self, root_inode, name=self.superblock.s_volume_name)

        # NOTE 1: When I calculate block_group_count, I round up to the upper integer (*) because apparently the free space
        #         at the end of the partition is assigned to one more 'block group', but which obviously does not respect having
        #         'superblock.s_blocks_per_group' blocks, but less.
        #         (anyway, this is by way of comment, as I did it works as I wanted, I don't think the logic needs to be changed)

        # NOTE 2: Maybe all the calculations referring to the number of block groups, should go in the class Superblock.
        #         (due to what I do in the __str__ and in the 2 lines prior to make the list of group descriptors*).

    def __repr__(self):
        return f"< ext2 @ {self.base_address} of {self.path}>\n"

    def __str__(self):
        # NOTE: The filesystem is divided into blocks as follows:
        #       boot_area + (total_blocks - 1 // blocks_per_group)*blocks_per_group + (total_blocks-1 % blocks_per_group)
        #       OR if block_size > 1K:
        #       (total_blocks // blocks_per_group)*blocks_per_group + (total_blocks % blocks_per_group)
        useful_blocks = self.superblock.s_blocks_count - self.superblock.s_first_data_block
        last_block_group_len = useful_blocks % self.superblock.s_blocks_per_group
        aux = f" (the last one has {last_block_group_len} blocks assigned)"
        if last_block_group_len == 0:
            # all block groups have fitted perfectly
            aux = ""

        return (
                f"< ext2 Filesystem:\n"
                f"{'Number of block groups:':49}{len(self.group_descriptors)}{aux}\n"
                f"{self.superblock}>\n"
            )

    def read_record(self, block_number, length, offset=0):
        """
        Method that reads a certain amount of bytes (length) from a given block
        of the filesystem (block_number), starting from an offset (by default 0).
        """
        block_size = self.superblock.s_log_block_size
        address = block_number*block_size             # This number will be multiple of 512 as 'block_size' is.
        self.handle.seek(self.base_address + address) # 'base_address' is also multiple, since it must be the sector of the device where the partition begins.
        for i in range(offset//block_size):           # I do this mostly for very large offset cases*
            self.handle.seek(block_size, 1)
        # here I have to do a 'read' to move the device's pointer, because if I did a seek, the next read could give an error for just not being positioned in a multiple of 512.
        self.handle.read(offset % block_size)         # *and then this read is small: it will be less than the block size (<= block_size-1),
        raw_record = self.handle.read(length)
        return raw_record

        # Testing, I discovered that there are some 'restrictions' when wanting
        # to read something from a storage device* (this is already outside the filesystem).
        # We can only read if we are positioned (seek) in a multiple of 512 bytes.
        # Once there, we make a first reading, and then we will have the next 8192 bytes
        # enabled/allowed to do some seek in that interval and perform a read without problems.
        # (Even if the read goes beyond that limit later, there is no problem with that).
        # Otherwise, an exception will be thrown when trying to read something.

        # Therefore, I replace a seek(offset) with a read(offset); it is not very nice but it works.
        
        # * NOTE: For 'normal' files, this problem does not arise, therefore we could analyze
        #         for example a disk image (.vhd) without worrying much about all this.

    def read_block(self, block_number):
        """
        Method that reads a block from the filesystem (block_number >= 0).
        """
        block_size = self.superblock.s_log_block_size
        address = block_number*block_size
        self.handle.seek(self.base_address + address) # here everything goes well, because it is guaranteed that we will be in a multiple of 512 bytes.
        raw_block = self.handle.read(block_size) # therefore this read does not give problems.
        return raw_block

        #block_size = self.superblock.s_log_block_size
        #self.handle.seek(self.base_address)
        #for i in range(block_number):
        #    self.handle.seek(block_size, 1)
        #raw_block = self.handle.read(block_size)
        #return raw_block

    def read_inode(self, inode_number):
        """
        Method that reads an inode from the filesystem (inode_number >= 1).
        Attention! Unlike block numbering, the inodes start to be numbered from 1,
        since inode number 0 indicates 'null inode' (so the first entry in
        the inode table 'inode_table[0]', corresponds to inode 1 and not inode 0).
        """
        if inode_number < 1:
            raise TypeError("the inode number must be >= 1 !")
        # locating the inode
        block_group = (inode_number - 1) // self.superblock.s_inodes_per_group
        first_inode_table_block = self.group_descriptors[block_group].bg_inode_table
        local_inode_index = (inode_number - 1) % self.superblock.s_inodes_per_group
        # and now we read it
        raw_inode = self.read_record(first_inode_table_block, self.INODE_STRUCT_SIZE, offset=local_inode_index*self.INODE_STRUCT_SIZE)
        return raw_inode

    def _parse_direct_blocks(self, pointers):
        """
        Method that parses a list of direct pointers to data blocks.
        We already have the number of each data block, but what I do is filter
        the pointers to null blocks (p-> 0).
        """
        data_block_numbers = [pointer for pointer in pointers if pointer != 0]
        return data_block_numbers
        # There may be unallocated blocks yet (pointers to 0), so we will only keep those pointers that do not equal to null (0).

    def _parse_indirect_1_block(self, block_number):
        """
        Method that parses the pointers contained in a simple indirect block,
        given its block number within the filesystem. By containing direct pointers,
        we directly obtain all the numbers of data block that indirectly points.
        """
        raw_indirect_1_block = self.read_block(block_number)
        indirect_1_block = [pointer[0] for pointer in struct.iter_unpack("<I", raw_indirect_1_block) if pointer[0] != 0]
        return indirect_1_block
        # Each pointer occupies 4 bytes, therefore each indirect block will have block_size/4 pointers to other blocks.

    def _parse_indirect_2_block(self, block_number):
        """
        Method that parses the pointers contained in a double indirect block,
        given its block number within the filesystem. By containing pointers
        that point to simple indirect blocks, we will use the '_parse_indirect_1_block'
        method for each of them, and thus obtain all the data block numbers
        that are pointed at the end of this indirection chain.
        """
        raw_indirect_2_block = self.read_block(block_number)
        indirect_2_block = [pointer[0] for pointer in struct.iter_unpack("<I", raw_indirect_2_block) if pointer[0] != 0]
        
        data_block_numbers = []

        for p in indirect_2_block:
            data_block_numbers += self._parse_indirect_1_block(p)

        return data_block_numbers

    def _parse_indirect_3_block(self, block_number):
        """
        Method that parses the pointers contained in a triple indirect block,
        given its block number within the filesystem. By containing pointers
        that point to double indirect blocks, we will use the '_parse_indirect_2_block'
        method for each of them, and thus obtain all the data block numbers
        that are pointed at the end of this indirection chain.
        """
        raw_indirect_3_block = self.read_block(block_number)
        indirect_3_block = [pointer[0] for pointer in struct.iter_unpack("<I", raw_indirect_3_block) if pointer[0] != 0]
        # how much power in a single line of code! it's beautiful, Python at its best!

        data_block_numbers = []

        for p in indirect_3_block:
            data_block_numbers += self._parse_indirect_2_block(p)

        return data_block_numbers


    def open(self, path):
        """
        Method that "opens" a file or directory on the filesystem.
        It returns an object of type "Directory" or "FileHandle" depending on
        what we are looking for (according to what we specify in the 'path',
        which must be absolute!).
        With the returned object, we can make use of its methods to see
        the content of a directory, or the data of a file, as appropriate
        (among other things).
        """
        path = path.encode(ENCODING) # ext2 is case sensitive, so I handle the path as it comes.
        parts = path.split(b'/')[1:] # we decompose the path in a list with the directory/file names.
        obj = self.root # we always start from the root directory.
        if path != b'/': # in case we want to open only the root.
            for name in parts: # We will go into directory by directory until we reach the file or directory we are looking for.
                try:
                    # we make a dictionary <name: inode> from the directory entries of the current directory.
                    files = {f.name:f.inode for f in obj.files}
                except AttributeError:
                    # in case something like this happens: /dir_1/file.txt/dir_2
                    # a file can only be at the end of the path, not in the middle (everything else must be directories).
                    raise FileNotFoundError(f"{obj.path} is not a Directory!") # we will see this exception by console.
                if name in files: # we verify that the searched directory/file name is in a directory entry of the current directory.
                    # We locate, read and parse the inode
                    inode_number = files[name]
                    raw_inode = self.read_inode(inode_number)
                    inode_obj = inode.Inode(raw_inode)
                    # and we instantiate the directory or file as of the inode that represents it.
                    if inode_obj.i_mode[0] == 'd': # Maybe this way of checking the file type needs to be improved.
                        obj = Directory(self, inode_obj, name, parent=obj)
                    else:
                        obj = FileHandle(self, inode_obj, name, parent=obj) # if we get here before finishing the 'for', the exception 'AttributeError' will be thrown above.
                else:
                    raise FileNotFoundError(f"No such file or directory {obj.path}/{name.decode(ENCODING)}")
            
        return obj

    # I don't know if it's the proper method name, but I wanted to put the 'close' here.
    def unmount(self):
        self.handle.close()
        return True



# To check:
#  - nothing else for now.

import struct

# Types of files recognized by Ext2
file_types = {
    0: "Unknown",
    1: "Regular file",
    2: "Directory",
    3: "Character device",
    4: "Block device",
    5: "Named pipe",
    6: "Socket",
    7: "Symbolic link"
}

# Filename encoding
ENCODING = 'latin-1' # generally latin-1 is used

class DirectoryEntry:
    """
    Class representing a Directory Entry of an ext2 filesystem.
    (are the records stored in the directory blocks pointed to by directory inodes)
    (I call them 'directory blocks', but they count as data blocks of the filesystem)

    The structure has a variable length, since the 'name' field is a
    variable length array of up to 255 characters. Moreover, the length of a
    directory entry is always a multiple of 4 and, therefore, null characters ('\0')
    are added for padding at the end of the filename, if necessary.
    (but the 'name_len' field stores the actual file name length)
    """
    def __init__(self, data=bytes(8),
                 inode=None, rec_len=None, name_len=None, file_type=None, name=None
                 ):

        # let's parse
        p_inode     = struct.unpack("<I", data[0:4])[0]
        p_rec_len   = struct.unpack("<H", data[4:6])[0]
        p_name_len  = struct.unpack("<B", data[6:7])[0]
        p_file_type = struct.unpack("<B", data[7:8])[0]
        # I directly assign the raw binaries to it and then in the '__str__' I make it a 'decode'.
        p_name      = data[8:8+p_name_len] # variable length
        # note: by default, the 'slice' will return b'' (empty bytestring) and not IndexError

        self._raw_data = data
        # and now we set, either the parameterized values or the parsed ones
        self.inode     = inode     or p_inode
        self.rec_len   = rec_len   or p_rec_len
        self.name_len  = name_len  or p_name_len
        self.file_type = file_type or p_file_type
        self.name      = name      or p_name

    @property
    def raw_data(self):
        """
        Bytes to be parsed.
        (are the [8 to n] bytes corresponding to the structure of a directory entry)
        """
        return self._raw_data

    @raw_data.setter
    def raw_data(self, value):
        pass
    
    # ---

    @property
    def inode(self):
        """
        Inode number representing the file or directory
        """
        return self._inode

    @inode.setter
    def inode(self, value):
        # This way we prevent someone from sending us data types that do not have
        # a numerical representation. That is, in those cases, an exception will be thrown.
        self._inode = int(value)

    @property
    def rec_len(self):
        """
        Directory entry length (in bytes)
        """
        return self._rec_len

    @rec_len.setter
    def rec_len(self, value):
        self._rec_len = int(value)

    @property
    def name_len(self):
        """
        Filename length
        """
        return self._name_len

    @name_len.setter
    def name_len(self, value):
        self._name_len = int(value)

    @property
    def file_type(self):
        """
        Returns a string corresponding to the file type
        """
        return file_types[self._file_type]

    @file_type.setter
    def file_type(self, value):
        value = int(value)
        # to avoid a KeyError
        if value < 0 or value > 7:
            value = 0
        self._file_type = value

    @property
    def name(self):
        """
        Filename (up to 255 chars)
        """
        return self._name

    @name.setter
    def name(self, value):
        # I admit both strings and bytestrings, but internally I keep bytestrings.
        if isinstance(value, str):
            value = bytes(value, ENCODING)
        self._name = value

    # ---

    def __str__(self):
        return (
                f"Inode number:           {self.inode}\n"
                f"Directory entry length: {self.rec_len}\n"
                f"Filename length:        {self.name_len}\n"
                f"File type:              {self.file_type}\n"
                f"Filename:               {self.name.decode(ENCODING)}\n"
            )




"""
Defines class and functions to handle GUIDs.
"""
import struct


from .utils import str2bytes


class GUID:
    """GUID class"""
    def __init__(self, data, *, mixed_endian=False):
        if isinstance(data, str):
            data = data.replace("{", "")
            data = data.replace("}", "")
            data = data.replace("-", "")
            data = str2bytes(data)
        self._data = data
        if mixed_endian:
            gp1, gp2, gp3 = struct.unpack("<LHH", data[0: 8])
        else:
            gp1, gp2, gp3 = struct.unpack(">LHH", data[0: 8])
        gp4, gp5, gp6 = struct.unpack(">HHL", data[8:16])
        gp5 = (gp5 << 32 ) | gp6
        self.gp1, self.gp2, self.gp3 = gp1, gp2, gp3
        self.gp4, self.gp5           = gp4, gp5
        self._repr_str  = f"{gp1:08x}-{gp2:04x}-{gp3:04x}-{gp4:04x}-{gp5:012x}".upper()
        self._mixed_endian = mixed_endian

    def __bytes__(self):
        if self._mixed_endian:
            data = struct.pack("<LHH", self.gp1, self.gp2, self.gp3)
        else:
            data = struct.pack(">LHH", self.gp1, self.gp2, self.gp3)
        gp5 = self.gp5 >> 32
        gp6 = self.gp5 & 0x0000ffffffff
        data += struct.pack(">HHL", self.gp4, gp5, gp6)
        return data

    def __repr__(self):
        return f"< GUID: {self._repr_str} >"
    
    def __str__(self):
        return "{" f"{self._repr_str}" "}"
    
    # and to make these dictionary-key-able...
    def __hash__(self):
        return hash(self._repr_str)
    
    def __eq__(self, other):
        if not isinstance(other, GUID):
            return False
        return self._repr_str == other._repr_str
    
    def __ne__(self, other):
        return not(self == other)



"""
This module handles reading from and writing to a FAT32 filesystem.
"""
import calendar
import datetime
import struct
import sys


from .utils import slicer


# TODO: read https://stackoverflow.com/questions/13775893/converting-struct-to-byte-and-back-to-struct
#       and research a bit into this being a reasonable way to convert bytes
#       into structs (and back).
# TODO: consider widening support for other encodings (eg: latin1, utf-8, etc)
#       in file/directory names.
# TODO: Linux-style LFN support (see https://lkml.org/lkml/2009/6/26/314)     


ATTRIBUTES = {
    "read-only": 0x01, 
    "hidden"   : 0x02,
    "system"   : 0x04,
    "volume-id": 0x08,
    "directory": 0x10,
    "archive"  : 0x20,
}

# TODO: this is actually a hack right now, and there might be a simpler way to
#       do it
ALL_ASCII = b"".maketrans(b"", b"")  # a bit hacky, but works
FILENAME_CHARS = b"0123456789 ABCDEFGHIJKLMNOPQRSTUVWXYZ!#$%&'()-@^_`{}~\xe5"
REMAIN_ASCII = ALL_ASCII
for idx, char in enumerate(FILENAME_CHARS):
    char = FILENAME_CHARS[idx:idx + 1]
    REMAIN_ASCII = REMAIN_ASCII.replace(char, b"")
FILENAME_TRANS = b"".maketrans(
    REMAIN_ASCII,
    b"\x00" * len(REMAIN_ASCII)
)
# TODO: the whole PATH_SEP thing might need better thought, but for the moment
#       simple should be enough
PATH_SEP = "\\"


def read_attributes(value):
    """
    Receives the raw attributes byte, and parses it into a dictionary based on
    the `ATTRIBUTES` constant.

    :param raw_attrs: bytes object, only the first byte is used
    :return: dictionary with the corresponding attributes and their bool value
    """
    ret = {}
    for k, v in ATTRIBUTES.items():
        ret[k] = bool(value & v)
    return ret


def read_time(bytes_, mili=0):
    """
    Receives time + date bytes in `bytes_`, and returns a datetime object.
    Checks for pathological 0xffff patterns.

    :param bytes_: raw value, made from the the time+date fields in the raw
        directory entry/file record.
    :param mili: the raw byte (int) that encodes the 10 milisecond precision,
        for timestamps that support it (defaults to 0)
    :return: datetime.datetime object
    """
    raw_time, raw_date, = struct.unpack("<HH", bytes_)
    # first we take care of the date
    year    = (raw_date >> 9) + 1980
    month   = (raw_date & 0b0000000111100000) >> 5
    day     =  raw_date & 0b0000000000011111
    # and now the time
    hour    =  raw_time >> 11
    minute  = (raw_time & 0b0000011111100000) >> 5
    second  = (raw_time & 0b0000000000011111) * 2
    second += mili // 100
    micros  = (mili % 100) * 10000
    # we know theres an issue in some Linux based systems that make
    # 0xffffffff datetimes for some FileRecords (that don't seem to belong to
    # the files, some kind of temporary record) so we must check a few things:
    year   = max(1980,  min(2107,    year))  # clips the values to within range
    month  = max(1,     min(12,     month))
    _weeks, max_days = calendar.monthrange(year, month)
    day    = max(1,     min(max_days, day))
    hour   = min(23,   hour)
    minute = min(59, minute)
    second = min(59, second)
    # that gives a few sanity checks and should catch that particular issue
    dt = datetime.datetime(year, month, day, hour, minute, second, micros)
    return dt


class FileRecord:
    """
    An entry inside a FAT32 directory.
    """
    def __init__(self, data):
        self._raw_data   = data
        self._name       = b""
        self._ext        = b""
        self._size       = -1
        self._attributes = {k: False for k in ATTRIBUTES}
        self._flags      = 0  # reserved, should be 0 but different implementations may use it 
        self._cluster    = -1
        self._created    = None
        self._last_access= None
        self._modified   = None
        # that's all for real attributes of a file record
        self._parse()
    
    # The Properties
    # (these could be done with metaclasses, but...)
    @property
    def raw_data(self):
        """
        The raw bytes used to initialize the File Record. Read only for now.
        """
        return self._raw_data
    
    @raw_data.setter
    def raw_data(self, value):
        pass  # let's make it read-only for the moment

    @property
    def name(self):
        """
        The name of the File Record.

        When setting this property, checks are made to make sure it can be
        represented as the packed structure on disk.
        """
        return self._name
    
    @name.setter
    def name(self, value):
        if isinstance(value, str):
            value = bytes(value, "latin1")
        value = value.upper()  # not supporting long names for the moment
        name = value.translate(FILENAME_TRANS)
        name = name.replace(b"\x00", b"")
        name = name[0:8]   # byebye long names!
        # let's check if there's an ext, for the lazy user
        if b"." in value:
            name, dot, ext = value.rpartition(b".")
            ext = ext.rstrip()
            self._ext = ext[-3:]
        name = name.rstrip()
        self._name = name # 
        # will not enforce name uniqueness, just being uppercase and within
        # valid range of characters
    
    @property
    def ext(self):
        """
        The extension of the File Record.

        When setting this property, checks are made to make sure it can be
        represented as the packed structure on disk.
        """
        return self._ext
    
    @ext.setter
    def ext(self, value):
        if isinstance(value, str):
            value = bytes(value, "latin1")
        value = value.upper()
        ext = value.translate(FILENAME_TRANS)
        ext = ext.replace(b"\x00", b"")
        ext = ext.rstrip()
        ext = ext[0:3]
        self._ext = ext
    
    @property
    def fullname(self):
        """
        Helper property, returns name for directories and name.ext for files.
        Read only.
        """
        if self.attributes["directory"]:
            return self.name
        return self.name + b"." + self.ext
    
    @fullname.setter
    def fullname(self, value):
        pass
    
    @property
    def size(self):
        """
        The size of the File Record.

        When setting this property, checks are made to make sure it can be
        represented as the packed structure on disk. Values larger than
        0xffffffff are truncated at it.
        """
        return self._size
    
    @size.setter
    def size(self, value):
        if value > 0xffffffff:
            value = 0xffffffff
        self._size = value
    
    @property
    def attributes(self):
        """
        The attributes of the File Record.

        When setting this property, checks are made to make sure it can be
        represented as the packed structure on disk.
        """
        return self._attributes
    
    @attributes.setter
    def attributes(self, value):
        # let's cleanup in case the given dict has some extra keys
        value = {k:v for k, v in value.items() if k in ATTRIBUTES}
        self._attributes.update(value)
    
    @property
    def flags(self):
        """
        The flags of the File Record. Read only.

        This is actually a reserved field, which non-Microsoft implementations
        use for different purposes.
        """
        return self._flags
    
    @flags.setter
    def flags(self, value):
        self._flags = value  # we're gonna need this for the optional arguments

    @property
    def cluster(self):
        """
        The first cluster of the File Record.

        When setting this property, checks are made to make sure it can be
        represented as the packed structure on disk. Same as with size, we
        simply check for the value to fit in a uint32. The filesystem should
        be responsible to check that the referenced cluster is valid and not
        out of range.
        """
        return self._cluster
    
    @cluster.setter
    def cluster(self, value):
        if value > 0xffffffff:
            value = 0xffffffff
        # no point in checking anything else, the filesystem should make
        # sure there's no invalid cluster set
        self._cluster = value
    
    @property
    def created(self):
        """
        The creation timestamp of the File Record.

        When setting this property, checks are made to make sure it can be
        represented as the packed structure on disk. Those are:

        * Year is set to be in range 1980-2107 (truncated at those values for
          smaller/larger).
        * Microsecond accuracy of the datetime object is capped to 10ms
          resolution (as can be represented in structure)
        """
        return self._created
    
    @created.setter
    def created(self, value):
        # truth is, python's datetime object handles a lot already
        # so we only need to check the year
        year = value.year
        if year < 1980:
            year = 1980
            value = value.replace(year=year)
        elif year > 2107:
            year = 2107
            value = value.replace(year=year)
        micros = value.microsecond
        micros = (micros // 10000) * 10000  # let's be honest about the precision
        value.replace(microsecond=micros)
        self._created = value

    # last_access
    @property
    def last_access(self):
        """
        The last accessed timestamp of the File Record.

        When setting this property, checks are made to make sure it can be
        represented as the packed structure on disk. Those are:

        * Year is set to be in range 1980-2107 (truncated at those values for
          smaller/larger).
        * Since the structure only has 2 bytes for a date, the datetimes object
          hour, minut, second and microsecond are set to 0.
        
        It is technically a datetime still to be consistent with the other
        timestamps.
        """
        return self._last_access
    
    @last_access.setter
    def last_access(self, value):
        # truth is, python's datetime object handles a lot already
        # so we only need to check the year
        year = value.year
        if year < 1980:
            year = 1980
            value = value.replace(year=year)
        elif year > 2107:
            year = 2107
            value = value.replace(year=year)
        # and here we erase the time
        value = value.replace(hour=0, minute=0, second=0, microsecond=0)
        # technically a date object at this point, but we prefer to keep
        # things consistent and have them all be datetime
        self._last_access = value

    @property
    def modified(self):
        """
        The modified timestamp of the File Record.

        When setting this property, checks are made to make sure it can be
        represented as the packed structure on disk. Those are:

        * Year is set to be in range 1980-2107 (truncated at those values for
          smaller/larger).
        * The datetimes object second resolution is cut to 2 second icnrementes
          (as the on-disk structure allows) and microseconds are set to 0.
        """
        return self._modified
    
    @modified.setter
    def modified(self, value):
        # truth is, python's datetime object handles a lot already
        # so we only need to check the year
        year = value.year
        if year < 1980:
            year = 1980
            value = value.replace(year=year)
        elif year > 2107:
            year = 2107
            value = value.replace(year=year)
        second = value.second
        second = (second // 2) * 2  # let's be honest about the precision
        value = value.replace(second=second, microsecond=0)
        self._modified = value
    
    @property
    def deleted(self):
        """
        Boolean for the FileRecord being marked as deleted.

        Setting this property to True (or True-like values) re-writes the first
        character the FileRecord's name attribute. Setting it to false checks
        for a \\xe5 value on the first byte, and if it finds it, replaces it for
        and underscore.
        """
        return self._name[0:1] == b"\xe5"
    
    @deleted.setter
    def deleted(self, value):
        if value:
            self._name = b"\xe5" + self._name[1:]
        else:
            if self._name[0:1] == b"\xe5":
                self._name = b"_" + self._name[1:]
    
    def __bytes__(self):
        return b""
    
    def __repr__(self):
        name = self.name.decode("latin1")
        ext  = self.ext.decode("latin1")
        if self.attributes["directory"]:
            showname = f"<DIR> {name}"
        else:
            showname = f"{name}.{ext}"
        return f"< DirectoryEntry: {showname}>"
    
    def __str__(self):
        name = self.name.decode("latin1")
        ext  = self.ext.decode("latin1")
        if self.attributes["directory"]:
            showname = f"<DIR> {name}"
        else:
            showname = f"{name}.{ext}"
        return (
               f"< DirectoryEntry: {showname}\n"
               f"    {'size':12}:{self.size:>12}\n" 
               f"    {'attributes':12}:{self._attrs2str():>12}\n" 
               f"    {'cluster':12}:{self.cluster:>12}\n" 
               f"    {'created':12}: {self.created}\n" 
               f"    {'last_access':12}: {self.last_access}\n" 
               f"    {'modified':12}: {self.modified}\n" 
               f"    {'deleted':12}:{str(self.deleted):>12}\n" 
               f">"
        )
    
    def _attrs2str(self):
        """
        Helper function for pretty-printing the attributes in __str__.
        """
        ret = [
            k[0].upper() if v else k[0]
            for (k, v) in self._attributes.items()
        ]
        return "".join(ret)
    
    def _parse(self):
        """
        Parses self._raw_data and updates the properties of the file.
        """
        # TODO: maybe it should be called load? or just parse?
        #       will define it better after having a dump() or save() method
        self.name        = self._raw_data[0:8] + b"." + self._raw_data[8:11]
        # we use the shorthand that sets name and extension in a single pass
        self.size,       = struct.unpack("<L", self._raw_data[28:32])
        self.attributes  = read_attributes(self._raw_data[11])
        self.flags       = self._raw_data[12]
        self.cluster,    = struct.unpack("<L", self._raw_data[26:28] + self._raw_data[20:22])
        self.created     = read_time(self._raw_data[14:18], mili=self._raw_data[13])
        self.last_access = read_time(b"\x00\x00" + self._raw_data[18:20])
        self.modified    = read_time(self._raw_data[22:26])



class Directory:
    """
    Class to handle the abstraction for reading (eventually writing/modifying)
    a directory/folder in a FAT32 filesystem. As with FileHande, this class is
    strange, because it's more involved than you'd usually find, and has to
    take care of some things which are usually responsibility of the OS.

    :param filesystem: the base filesystem (of class FAT32) to which the
        Directory belongs
    :param record: the FileRecord (Directory Entry) for the directory
    :param parent: (Directory) parent of the (new) directory, used to build a
        path string
    :return: Directory object in `filesystem `for the given `record`
    """
    def __init__(self, filesystem, record, *, parent=None):
        self._filesystem = filesystem
        if record is None:
            record = FileRecord(bytes(32))
            record.cluster = filesystem.root_cluster
        self._record = record
        self.cluster = record.cluster
        # a bit of a hack for the root cluster
        self._parent     = parent
        self.files       = []
        if parent is None:
            self.path = record.name.decode("latin1")
        else:
            self.path = parent.path + PATH_SEP + record.name.decode("latin1")
        self._parse()
    
    def __repr__(self):
        return f"< Directory {self.path}>"
    
    def __str__(self):
        ret =  f"< Directory {self.path}\n"
        ret += "  Files: [\n    "
        ret += "\n    ".join([repr(f) for f in self.files])
        ret += "\n  ]"
        ret += "\n>"
        return ret
    
    def _parse(self):
        """
        Parses the directory cluster on disk, creating the FileRecords for every
        entry in the directory (wink). For the moment long file names (LFNs) are
        skipped since we don't want to tread into Microsoft patents territory.
        Might think of supporting Linux-stlye LFNs.
        """
        
        def lfn_filter(record):
            """
            Returns false for records which are LFN data.
            """
            attrs = record.attributes
            lfn = (
                attrs["read-only"] and
                attrs["hidden"] and
                attrs["system"] and
                attrs["volume-id"]
            )
            return not(lfn)

        filesystem   = self._filesystem
        record       = self._record
        raw_data     = []
        cluster      = self.cluster
        next_cluster = filesystem.fat1[cluster]
        while next_cluster <= 0x0ffffff0:
            raw_data.append(filesystem._read_cluster(cluster))
            cluster = next_cluster
            next_cluster = filesystem.fat1[cluster]
        raw_data.append(filesystem._read_cluster(cluster))
        raw_data  = b"".join(raw_data)
        self.files = list(filter(
            lfn_filter,
            (
                FileRecord(s)
                for s in slicer(raw_data, 32)
                if s != bytes(32)
            )
        ))

class FileHandle:
    """
    Class to handle reading (and eventually writing) files in the FAT32
    filesystem. This class is a lot more involved than pythons standard file
    object, because it actually deals with the filesystem on a level closer to
    the OS.
    """
    def __init__(self, filesystem, record, mode="rb", *, parent=None):
        self._filesystem = filesystem
        self._record     = record
        self._mode       = mode
        if parent is None:
            self.path = (record.name + b"." + record.ext).decode("latin1")
        else:
            self.path = parent.path + PATH_SEP + (record.name + b"." + record.ext).decode("latin1")
        # this is all a lie (yet), we'll just open in "rb" mode
        # we will try to mimic the IOBase methods as close as possible, without
        # going crazy in it... right?
        self.closed       = False
        # and we'll need a few other attributes as well
        self._buffer      = bytearray(filesystem._read_cluster(record.cluster))
        self._buffer_pos  = 0  # for easier handling of the buffer and _file_pos
        self._buffer_size = filesystem.sectors_per_cluster * filesystem.bytes_per_sector
            # we buffer a cluster at a time
        self._ccluster    = record.cluster
        # we keep track of which cluster we're own to read the next one when
        # necessary
        self._file_pos    = 0  # position in the file, .tell() returns this
        self._readable    = True  # will set along with mode, when supported
    
    def __repr__(self):
        return f"< FileHandle for {self.path}>"
    
    def __str__(self):
        return self.__repr__()

    def __del__(self):
        # should probably do something about writing, for example
        # and flusihing changes to disk
        # once writing is supported, that is
        pass
    
    def close(self):
        """
        Closes the file.
        """
        self.closed = True
    
    def fileno(self):
        pass

    def flush(self):
        pass

    def isatty(self):
        return False
    
    def read(self, size=-1):
        bsize = self._buffer_size
        start = self._buffer_pos
        if size < 0:
            size = self._record.size - self._file_pos
        ret = []
        # and now we read
        while (bsize - start) < size:  # we have to read clusters from the chain
            ret.append(self._buffer[start:])
            size -= (bsize - start)
            self._file_pos += (bsize - start)
            ncluster = self._filesystem.fat1[self._ccluster]
            if ncluster >= 0x0ffffff0:
                # magic EOF cluster number, found by trial and error
                # TODO: check that the cluster is within bounds of the fs
                break
            start = 0
            self._ccluster = ncluster
            self._buffer = bytearray(self._filesystem._read_cluster(ncluster))
        span = min(self._record.size - self._file_pos, size)
        ret.append(self._buffer[start: start + span])
        self._buffer_pos =  start + span
        self._file_pos   += span
        return b"".join(ret)
    
    def readable(self):
        """
        Returns whether the file is readable or not.
        """
        return self._readable
    
    def readlines(self, hint=-1):
        pass  # not ready to implement this yet

    def seek(self, offset, whence=0):
        """
        Moves the file pointer to offset.

        :param offset: position to move the file pointer
        :param whence: not implemented (yet)
        :return: the current position of the file pointer, after moving (will
            be equal to `offset`, unless something goes very wrong)
        """
        filesystem = self._filesystem
        record     = self._record
        bsize      = self._buffer_size
        # it's not pretty but it's simple -- we'll have to do a scan of the
        # FAT to see in which cluster we have to land, and buffer that
        # c_idx will be our cluster index
        c_idx = offset // bsize
        self._ccluster = record.cluster
        ncluster = self._filesystem.fat1[self._ccluster]
        while (ncluster < 0x0ffffff0) and c_idx > 0:
            self._ccluster = ncluster
            c_idx -= 1
            ncluster = self._filesystem.fat1[self._ccluster]
        if offset >= record.size:  # we went over the end of the file
            self._buffer     = bytearray(filesystem._read_cluster(self._ccluster))
            self._file_pos   = record.size
            self._buffer_pos = record.size % bsize
            return self._file_pos
        self._buffer     = bytearray(filesystem._read_cluster(self._ccluster))
        self._buffer_pos = offset % bsize
        self._file_pos   = offset
        return self._file_pos
    
    def seekable(self):
        return True  # hardcoded for now, can't think a situation where it won't be
    
    def tell(self):
        """
        Returns the current position of the file pointer.
        """
        return self._file_pos
    
    def truncate(self, size=None):
        pass
        # will think about implementing this
    
    def writeable(self):
        return False  # hardcoded, will work on this later
    
    def writelines(self, lines):
        pass  # dummy method, for the moment


class FAT32:
    """
    Class that gathers methods and objects to handle a FAT32 filesystem.
    """
    def __init__(self, path, base_address=0):
        self.path          = path
        self._base_address = base_address
        self._handle       = open(path, "rb+")
        self._handle.seek(base_address)
        # and now for the header data...}
        buffer = self._handle.read(512)
        bps, spc, rs, nof = struct.unpack("<HBHB", buffer[0x0b:0x11])
        spf, _, rc = struct.unpack("<LLL", buffer[0x24:0x30])
        self.bytes_per_sector    = bps
        self.sectors_per_cluster = spc
        self.reserved_sectors    = rs
        self.number_of_fats      = nof
        self.sectors_per_fat     = spf
        self.root_cluster        = rc
        self.root                = None  # we set its value later
        # and now some calculations...
        self.fat1_address = rs * bps
        self.fat2_address = (rs + spf) * bps
        self.base_cluster_address = (rs + (nof * spf)) * bps
        # for all these addresses we have to add the base_address of the
        # filesystem, otherwise we'll read anywhere on the device/image
        self.fat1_address += base_address
        self.fat2_address += base_address
        self.base_cluster_address += base_address
        # and now we're set!
        # I should probably set these to be read only through properties...
        self.fat1 = []
        self.fat2 = []
        self._load_fats()
        self._post_init()
    
    def __repr__(self):
        return f"< FAT32 @ {self._base_address} of {self.path}>"
    
    def __str__(self):
        return (
               "< FAT32 Filesystem:\n"
               f"    {'base address':24}: {self._base_address}\n"
               f"    {'sectors per cluster':24}: {self.sectors_per_cluster}\n"
               f"    {'reserved sectors':24}: {self.reserved_sectors}\n"
               f"    {'number of FATs':24}: {self.number_of_fats}\n"
               f"    {'sectors per FAT':24}: {self.sectors_per_fat}\n"
               f"    {'root cluster':24}: {self.root_cluster}\n"
               f"    {'FAT1 address':24}: {self.fat1_address}\n"
               f"    {'FAT2 address':24}: {self.fat2_address}\n"
               f"    {'base cluster address':24}: {self.base_cluster_address}\n"
               f">"
        )
    
    def _load_fats(self):
        """
        Reads the FATs from disk (hardcoded to 2 FATs, because there should
        only be two of them).
        """
        spf = self.sectors_per_fat
        bps = self.bytes_per_sector
        self._handle.seek(self.fat1_address)
        raw_fat = self._handle.read(spf * bps)
        self.fat1 = [ v[0] for v in struct.iter_unpack("<L", raw_fat) ]
        # since we're literally where the second FAT starts, we can just read
        # from here on
        raw_fat = self._handle.read(spf * bps)
        self.fat2 = [ v[0] for v in struct.iter_unpack("<L", raw_fat) ]
    
    def _cluster_address(self, cluster):
        """
        Calculates `cluster` address in the file the filesystem is being read
        from.

        :param cluster: cluster number to get the address of
        :return: offset in base file at which the cluster starts
        """
        cluster -= 2
        bps = self.bytes_per_sector
        spc = self.sectors_per_cluster
        bca = self.base_cluster_address
        return bca + (cluster * spc * bps)
    
    def _read_cluster(self, cluster, nclusters=1):
        """
        Reads `nclusters` (defaults to 1) from the filesystem, starting at
        position `cluster` (in clusters, the final address is converted 
        internally).

        :param cluster: cluster number to start reading from
        :param nclusters: amount of clusters to read (defaults to 1)
        :return: bytes read from the underlying file/device
        """
        bps = self.bytes_per_sector
        spc = self.sectors_per_cluster
        bca = self.base_cluster_address
        # the first actual cluster is #2 (there are no #0 and #1) so we have to
        # adjust the cluster number we're going to read
        cluster -= 2
        pos = bca + (cluster * spc * bps)
        length = nclusters * bps * spc
        self._handle.seek(pos)
        return self._handle.read(length)
    
    def _post_init(self):
        """
        Finishes setting up values in the filesystem. Everything that comes here
        assumes the basic functionality of the filesystem is already working
        and adds ease-of-use things.
        """
        root = Directory(self, None)
        v_id = root.files[0]
        if v_id.attributes["volume-id"]:
            root.path = (v_id.name + v_id.ext).decode("ascii")
        self.root = root
    
    def open(self, path):
        """
        Opens a file or directory found in the absolute path passed.

        :param path: absolute path to open.
        :returns: FileHandle or Directory object, depending on the path.
        """
        path  = path.upper()
        path  = path.encode("latin-1")  # TODO: make the encoding an attribute...
        path  = path.replace(b"\\", b"/")
        parts = path.split(b"/")[1:]
        # we ignore the first one, assuming it's the root -- remember, these
        # should be absolute paths!
        obj = self.root
        for part in parts:
            try:
                files = obj.files
            except AttributeError:
                raise FileNotFoundError(f"{obj.path} is not a Directory!")
            try:
                idx = next( idx
                            for idx, f in enumerate(files)
                            if f.fullname == part
                )
            except StopIteration:
                raise FileNotFoundError(f"No such file or directory {obj.path}\\{part}")
            # if we're here, we got a a directory
            record = obj.files[idx]
            if record.attributes["directory"]:
                obj = Directory(self, record, parent=obj)
            else:
                obj = FileHandle(self, record, parent=obj)
                # hopefully this happens just once, otherwise 
        # we could safely assume our object is just fine...
        return obj



# To check:
#  - decide whether or not to show the attributes 'bg_pad' and 'bg_reserved' in '__str__'.
#  - In an 'ext2/3' book, this section of the group descriptor also talks about
#    data block and inode bitmaps, but I think I'll touch on that when I make
#    the 'ext2' class that handles everything. EDIT: it was not necessary to parse them.

import struct

class GroupDescriptor:
    """
    Class representing a Group Descriptor of an ext2 filesystem.
    Each block group has its own group descriptor (which has information about the group).

    Actually, each block group contains a table of group descriptors,
    that is, each entry contains the information of each block group in the filesystem.
    (yes, just like the superblock, there is a copy of this table in each block group)
    """
    def __init__(self, data=bytes(32),
                 bg_block_bitmap=None, bg_inode_bitmap=None, bg_inode_table=None,
                 bg_free_blocks_count=None, bg_free_inodes_count=None,
                 bg_used_dirs_count=None, bg_pad=None, bg_reserved=None
                 ):
        
        p_bg_block_bitmap      = struct.unpack("<I", data[0:4])[0]
        p_bg_inode_bitmap      = struct.unpack("<I", data[4:8])[0]
        p_bg_inode_table       = struct.unpack("<I", data[8:12])[0]
        p_bg_free_blocks_count = struct.unpack("<H", data[12:14])[0]
        p_bg_free_inodes_count = struct.unpack("<H", data[14:16])[0]
        p_bg_used_dirs_count   = struct.unpack("<H", data[16:18])[0]
        p_bg_pad               = struct.unpack("<H", data[18:20])[0]
        p_bg_reserved          = struct.unpack("<III", data[20:32])

        self._raw_data = data

        self.bg_block_bitmap      = bg_block_bitmap      or p_bg_block_bitmap
        self.bg_inode_bitmap      = bg_inode_bitmap      or p_bg_inode_bitmap
        self.bg_inode_table       = bg_inode_table       or p_bg_inode_table
        self.bg_free_blocks_count = bg_free_blocks_count or p_bg_free_blocks_count
        self.bg_free_inodes_count = bg_free_inodes_count or p_bg_free_inodes_count
        self.bg_used_dirs_count   = bg_used_dirs_count   or p_bg_used_dirs_count
        self.bg_pad               = bg_pad               or p_bg_pad
        self.bg_reserved          = bg_reserved          or p_bg_reserved

        # 'bg_': block group ; 'p_': parsed

    @property
    def raw_data(self):
        """
        Bytes to be parsed.
        (are the 32 bytes corresponding to the structure of a group descriptor)
        """
        return self._raw_data

    @raw_data.setter
    def raw_data(self, value):
        pass

    # ---

    @property
    def bg_block_bitmap(self):
        """
        Block number of block bitmap.
        """
        return self._bg_block_bitmap

    @bg_block_bitmap.setter
    def bg_block_bitmap(self, value):
        self._bg_block_bitmap = int(value)

    @property
    def bg_inode_bitmap(self):
        """
        Block number of inode bitmap.
        """
        return self._bg_inode_bitmap

    @bg_inode_bitmap.setter
    def bg_inode_bitmap(self, value):
        self._bg_inode_bitmap = int(value)

    @property
    def bg_inode_table(self):
        """
        Block number of first inode table block.
        """
        return self._bg_inode_table

    @bg_inode_table.setter
    def bg_inode_table(self, value):
        self._bg_inode_table = int(value)

    @property
    def bg_free_blocks_count(self):
        """
        Number of free blocks in the group.
        """
        return self._bg_free_blocks_count

    @bg_free_blocks_count.setter
    def bg_free_blocks_count(self, value):
        self._bg_free_blocks_count = int(value)

    @property
    def bg_free_inodes_count(self):
        """
        Number of free inodes in the group.
        """
        return self._bg_free_inodes_count

    @bg_free_inodes_count.setter
    def bg_free_inodes_count(self, value):
        self._bg_free_inodes_count = int(value)

    @property
    def bg_used_dirs_count(self):
        """
        Number of directories in the group.
        """
        return self._bg_used_dirs_count

    @bg_used_dirs_count.setter
    def bg_used_dirs_count(self, value):
        self._bg_used_dirs_count = int(value)

    @property
    def bg_pad(self):
        """
        16bit value used for padding the structure on a 32bit boundary.
        (Alignment to word)
        """
        return self._bg_pad

    @bg_pad.setter
    def bg_pad(self, value):
        self._bg_pad = value

    @property
    def bg_reserved(self):
        """
        12 bytes of reserved space for future revisions.
        (Nulls to pad out 32 bytes)
        """
        return self._bg_reserved

    @bg_reserved.setter
    def bg_reserved(self, value):
        self._bg_reserved = value

    # ---

    def __str__(self):
        return (
                f"Block number of block bitmap:            {self.bg_block_bitmap}\n"
                f"Block number of inode bitmap:            {self.bg_inode_bitmap}\n"
                f"Block number of first inode table block: {self.bg_inode_table}\n"
                f"Number of free blocks in the group:      {self.bg_free_blocks_count}\n"
                f"Number of free inodes in the group:      {self.bg_free_inodes_count}\n"
                f"Number of directories in the group:      {self.bg_used_dirs_count}\n"
                f"Alignment to word:                       {self.bg_pad}\n"
                f"Nulls to pad out 32 bytes:               {self.bg_reserved}\n"
            )




# To check:
#  - I'm not sure whether to do certain parse on the 'setter' or the 'getter' (like the case of 'i_mode' or 'i_flags')
#    -> FOR NOW, LET THEY REMAIN AS THEY ARE CURRENTLY.
#  - check if the inode functions (defined at the end of the class) have to be implemented somewhere.
#  - there is a small "issue" with the timestamps of files, because when doing a test on a pendrive that I formatted in ext2,
#    when reading any date/time, I realized that it was recorded/written in UTC-3 (my time zone),
#    and then the 'datetime.fromtimestamp()' method, when converting it to our time zone (UTC-3), it subtracts 3 hours,
#    then date is inconsistent. If when manipulating files on ext2, the dates are always written in the PC's time zone,
#    I have to add a parameter to the '.fromtimestamp()' method, so that it does not convert the date.
#    But the point is that when showing the dates, we will not know what time zone they belong to,
#    unless we know exactly the time zone of the PC where those dates were written.
#    WHATEVER CRITERION WE CHOOSE, APPLY THE SAME TO THE 'superblock.py' DATES.

import datetime
import struct

# Types of files recognized by Ext2 (used in 'i_mode' field)
file_types = {
    0x0: 'u', # "Unknown"
    0x1: 'p', # "Named pipe / FIFO"
    0x2: 'c', # "Character device"
    0x4: 'd', # "Directory"
    0x6: 'b', # "Block device"
    0x8: '-', # "Regular file"
    0xA: 'l', # "Symbolic link"
    0xC: 's', # "Socket"
}

# Permissions a user has on a file (used in 'i_mode' field)
access_rights = {
    0: '-', # denied  (b: 000)
    1: 'x', # execute (b: 001)
    2: 'w', # write   (b: 010)
    4: 'r', # read    (b: 100)
}

N_FLAGS = 14 # in case in the future I parse more (max 32)

# Flags indicate how ext2 should behave when accessing data pointed to by an inode.
file_flags = {
    # 0x0000: "", # (I was thinking of using this if I parsed it only as a string, and not as a list)
    # ---
    0x0001: "secure deletion",
    0x0002: "record for undelete",
    0x0004: "compressed file",
    0x0008: "synchronous updates",
    0x0010: "immutable file",
    0x0020: "append only",
    0x0040: "do not dump/delete file",
    0x0080: "do not update .i_atime",
    # -- Reserved for compression usage --
    0x0100: "dirty (modified)",
    0x0200: "compressed blocks",
    0x0400: "access raw compressed data",
    0x0800: "compression error",
    # ---
    0x1000: "b-tree format directory, hash indexed directory", # both have the same bit
    0x2000: "AFS directory"
    # 0x4000: "journal file data" # reserved for ext3
}

class Inode:
    """
    Class representing an i-node of an ext2 filesystem.
    An inode is associated with a file (or directory), and contains metadata
    about it and pointers to its assigned data blocks (or directory blocks).

    An inode corresponds to a single file, therefore there is one inode
    for each file (or directory) in the filesystem.

    All inodes have the same size: at least 128 bytes (which is what I parse; the fundamental part of an inode).
      if revision of ext2 = 0 -> inode_size = 128 bytes (always)
      if revision of ext2 > 0 -> 128 bytes <= inode_size <= block_size, and is a perfect power of 2
    (the superblock is who knows the value of inode_size, block_size and the revision of ext2)
    """
    def __init__(self, data=bytes(128),
                 i_mode=None, i_uid=None, i_size=None, i_atime=None, i_ctime=None,
                 i_mtime=None, i_dtime=None, i_gid=None, i_links_count=None,
                 i_blocks=None, i_flags=None, osd1=None, i_block=None, i_generation=None,
                 i_file_acl=None, i_dir_acl=None, i_faddr=None, osd2=None
                 ):

        p_i_mode        = struct.unpack("<H", data[0:2])[0]
        p_i_uid         = struct.unpack("<H", data[2:4])[0]
        p_i_size        = struct.unpack("<I", data[4:8])[0]
        p_i_atime       = struct.unpack("<I", data[8:12])[0]
        p_i_ctime       = struct.unpack("<I", data[12:16])[0]
        p_i_mtime       = struct.unpack("<I", data[16:20])[0]
        p_i_dtime       = struct.unpack("<I", data[20:24])[0]
        p_i_gid         = struct.unpack("<H", data[24:26])[0]
        p_i_links_count = struct.unpack("<H", data[26:28])[0]
        p_i_blocks      = struct.unpack("<I", data[28:32])[0]
        p_i_flags       = struct.unpack("<I", data[32:36])[0]
        p_osd1          = struct.unpack("<I", data[36:40])[0]
        p_i_block       = struct.unpack("<IIIIIIIIIIIIIII", data[40:100])
        p_i_generation  = struct.unpack("<I", data[100:104])[0]
        p_i_file_acl    = struct.unpack("<I", data[104:108])[0]
        p_i_dir_acl     = struct.unpack("<I", data[108:112])[0]
        p_i_faddr       = struct.unpack("<I", data[112:116])[0]
        p_osd2          = struct.unpack("<III", data[116:128])

        self._raw_data = data

        self.i_mode        = i_mode        or p_i_mode
        self.i_uid         = i_uid         or p_i_uid
        self.i_size        = i_size        or p_i_size
        self.i_atime       = i_atime       or p_i_atime
        self.i_ctime       = i_ctime       or p_i_ctime
        self.i_mtime       = i_mtime       or p_i_mtime
        self.i_dtime       = i_dtime       or p_i_dtime
        self.i_gid         = i_gid         or p_i_gid
        self.i_links_count = i_links_count or p_i_links_count
        self.i_blocks      = i_blocks      or p_i_blocks
        self.i_flags       = i_flags       or p_i_flags
        self.osd1          = osd1          or p_osd1
        self.i_block       = i_block       or p_i_block
        self.i_generation  = i_generation  or p_i_generation
        self.i_file_acl    = i_file_acl    or p_i_file_acl
        self.i_dir_acl     = i_dir_acl     or p_i_dir_acl
        self.i_faddr       = i_faddr       or p_i_faddr
        self.osd2          = osd2          or p_osd2

        # 'i_': inode ; 'p_': parsed

    @property
    def raw_data(self):
        """
        Bytes to be parsed.
        (are the 128 bytes corresponding to the base structure of an inode)
        """
        return self._raw_data

    @raw_data.setter
    def raw_data(self, value):
        pass

    # ---

    @property
    def i_mode(self):
        """
        File type and access rights (16 bits)
        The top 4 bits for the type, and the bottom 12 bits for the access rights.

        Returns a formatted-string (based on how it actually shows on Linux)
        """
        f_type   = self._i_mode >> 12
        permissions = f"{file_types[f_type]}"

        f_rights = self._i_mode & 0x0fff

        # process control bits (I don't really concatenate them in the final string, for now)
        set_process_user_id =  ((f_rights & 0b100000000000) >> 11) == 1
        set_process_group_id = ((f_rights & 0b010000000000) >> 10) == 1
        sticky_bit =           ((f_rights & 0b001000000000) >>  9) == 1

        # access rights bits
        owner_rights = (f_rights & 0b000111000000) >> 6
        group_rights = (f_rights & 0b000000111000) >> 3
        other_rights = (f_rights & 0b000000000111)

        rights = [owner_rights, group_rights, other_rights]

        for r in rights:
            permissions += access_rights[r & 0b100] # each bit has a meaning
            permissions += access_rights[r & 0b010]
            permissions += access_rights[r & 0b001]

        # permissions += f"{set_process_user_id} {set_process_group_id} {sticky_bit}"

        return permissions

    @i_mode.setter
    def i_mode(self, value):
        self._i_mode = int(value)

    @property
    def i_uid(self):
        """
        Owner identifier
        """
        return self._i_uid

    @i_uid.setter
    def i_uid(self, value):
        self._i_uid = value

    @property
    def i_size(self):
        """
        Effective length of the file in bytes
        """
        return self._i_size

    @i_size.setter
    def i_size(self, value):
        self._i_size = int(value)

    # ---------------------------------------------------------

    # All 'timestamps' of files in ext2 are based on 'POSIX time' (https://en.wikipedia.org/wiki/Unix_time),
    # that is, the number of seconds that have elapsed since the Unix epoch
    # (january 1st 1970, 00:00:00 UTC). All the fields that store dates in ext2,
    # have 4 bytes (unsigned), therefore 2^32 seconds can be stored (it will be enough until the year 2106 :-) )
    # Python has a method of the class 'datetime' (datetime module) that converts
    # 'POSIX time' to local time (it converts it to our time zone, not to UTC 0, which comes in handy)
    # See: https://docs.python.org/3/library/datetime.html#datetime.datetime.fromtimestamp
    # So, I will represent the timestamps with DateTime objects.

    @property
    def i_atime(self):
        """
        Last access timestamp
        """
        # https://docs.python.org/3/library/datetime.html#datetime.datetime.timestamp
        ts = self._i_atime if self._i_atime.timestamp() > 0. else "Not defined" # if seconds elapsed since Unix epoch > 0.
        return ts

    @i_atime.setter
    def i_atime(self, value):
        # value: seconds elapsed since Unix epoch
        #self._i_atime = datetime.datetime.fromtimestamp(value) # DateTime object
        self._i_atime = datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)
        # last edit: for now I keep the date as it comes, I don't convert it to the current PC time zone.

        # > According to what I researched, the dates should be written to the filesystem in UTC+0,
        #   but on testing, they are written according to the computer's time zone, that is,
        #   instead of converting the PC's time zone to UTC+0 and then do the subtraction
        #   'system_datetime_utc0 - unix_epoch' to obtain the resulting seconds,
        #   the subtraction is directly done with the current date of the PC.

    @property
    def i_ctime(self):
        """
        Creation timestamp
        """
        ts = self._i_ctime if self._i_ctime.timestamp() > 0. else "Not defined"
        return ts

    @i_ctime.setter
    def i_ctime(self, value):
        #self._i_ctime = datetime.datetime.fromtimestamp(value)
        self._i_ctime = datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)

    @property
    def i_mtime(self):
        """
        Last modification timestamp
        """
        ts = self._i_mtime if self._i_mtime.timestamp() > 0. else "Not defined"
        return ts

    @i_mtime.setter
    def i_mtime(self, value):
        #self._i_mtime = datetime.datetime.fromtimestamp(value)
        self._i_mtime = datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)

    @property
    def i_dtime(self):
        """
        Deletion timestamp        
        """
        ts = self._i_dtime if self._i_dtime.timestamp() > 0. else "Not defined"
        return ts

    @i_dtime.setter
    def i_dtime(self, value):
        #self._i_dtime = datetime.datetime.fromtimestamp(value)
        self._i_dtime = datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)

    # ---------------------------------------------------------

    @property
    def i_gid(self):
        """
        Group identifier
        """
        return self._i_gid

    @i_gid.setter
    def i_gid(self, value):
        self._i_gid = value

    @property
    def i_links_count(self):
        """
        Hard links counter
        """
        return self._i_links_count

    @i_links_count.setter
    def i_links_count(self, value):
        self._i_links_count = int(value)

    @property
    def i_blocks(self):
        """
        Number of data blocks (in units of 512 bytes) that have been allocated
        to the file (count of disk sectors).
        """
        return self._i_blocks

    @i_blocks.setter
    def i_blocks(self, value):
        self._i_blocks = int(value)

    @property
    def i_flags(self):
        """
        Returns a formated-string with the flags of the file
        """
        flag_list = []
        for b in range(0, N_FLAGS):         # I parse (for now) only 'n' bits of the 32 that "_i_flags" has.
            k = self._i_flags & (0b1 << b)  # I am applying masks of the type 1, 10, 100, 1000 ... (bits).
            if k != 0:
                flag_list.append(file_flags[k]) # the bits 0s will represent an empty string, and the 1s, a flag.
        return ", ".join(flag_list) # I do this to make the string look "better".
        # I'm not sure if this could be done with a "List Comprehension".

    @i_flags.setter
    def i_flags(self, value):
        self._i_flags = int(value)

    @property
    def osd1(self):
        """
        Specific operating system information (4 bytes)
        """
        return self._osd1

    @osd1.setter
    def osd1(self, value):
        self._osd1 = value

    @property
    def i_block(self):
        """
        Pointers to data blocks (12 direct and 3 indirect). I will store them in a list.
        Pointers to 0 are equal to null pointers, which means there is no block assigned/allocated yet.
        Otherwise, the value stored in the pointer is the block number that it refers to.
        """
        return self._i_block

    @i_block.setter
    def i_block(self, value):
        self._i_block = list(value)

    @property
    def i_generation(self):
        """
        File version (used when the file is accessed by a network filesystem)
        """
        return self._i_generation

    @i_generation.setter
    def i_generation(self, value):
        self._i_generation = value

    @property
    def i_file_acl(self):
        """
        File access control list
        """
        return self._i_file_acl

    @i_file_acl.setter
    def i_file_acl(self, value):
        self._i_file_acl = value

    @property
    def i_dir_acl(self):
        """
        Directory access control list (is not used for regular files)
        """
        return self._i_dir_acl

    @i_dir_acl.setter
    def i_dir_acl(self, value):
        self._i_dir_acl = value

    @property
    def i_faddr(self):
        """
        Fragment address
        """
        return self._i_faddr

    @i_faddr.setter
    def i_faddr(self, value):
        self._i_faddr = value

    @property
    def osd2(self):
        """
        Specific operating system information (12 bytes)
        """
        return self._osd2

    @osd2.setter
    def osd2(self, value):
        self._osd2 = value

    def __str__(self):
        return (
                f"File type and access rights:          {self.i_mode}\n"
                f"Owner identifier:                     {self.i_uid}\n"
                f"File length in bytes:                 {self.i_size}\n"
                f"Time of last file access:             {self.i_atime}\n"
                f"Time that inode last changed:         {self.i_ctime}\n"
                f"Time that file contents last changed: {self.i_mtime}\n"
                f"Time of file deletion:                {self.i_dtime}\n"
                f"Group identifier:                     {self.i_gid}\n"
                f"Hard links counter:                   {self.i_links_count}\n"
                f"Number of data blocks of the file:    {self.i_blocks} (in units of 512 bytes)\n"
                f"File flags:                           {self.i_flags}\n"
                f"Direct pointers to data blocks:       {self.i_block[0:12]}\n"
                f"Pointer to simple indirect block:     {self.i_block[12]}\n"
                f"Pointer to doubly-indirect block:     {self.i_block[13]}\n"
                f"Pointer to triply-indirect block:     {self.i_block[14]}\n"
                # some fields would be missing (which for the moment I decided not to show them).
            )

    # Ext2 Inode Operations

    # only if the inode refers to a regular file
    def ext2_truncate():
        pass

    # only if the inode refers to a directory
    def ext2_create():
        pass

    def ext2_lookup():
        pass

    def ext2_link():
        pass

    def ext2_unlink():
        pass

    def ext2_symlink():
        pass

    def ext2_mkdir():
        pass

    def ext2_rmdir():
        pass

    def ext2_mknod():
        pass

    def ext2_rename():
        pass

    # only if the inode refers to a symbolic link that can be fully stored inside the inode itself
    def ext2_readlink():
        pass

    def ext2_follow_link():
        pass

    # Note: If the inode refers to a character device file, to a block device file,
    #       or to a named pipe, the inode operations do not depend on the filesystem.




# To check:
#  - decide if the 'getter' thing should go in the 'setter' (in s_log_block_size and s_state).
#    -> FOR NOW LET EVERYTHING STAY AS IT IS NOW
#  - check if the superblock's own methods/operations (defined last) should be implemented somewhere.
#  - what I commented in 's_uuid.setter'.
#  - maybe it would be necessary to implement the getter of 's_feature_compat',
#    's_feature_incompat' and 's_feature_ro_compat' (https://www.nongnu.org/ext2-doc/ext2.html#s-feature-compat).
#  - maybe it would be necessary to implement the getter of 's_algo_bitmap' (https://www.nongnu.org/ext2-doc/ext2.html#s-algo-bitmap).
#  - maybe it would be nice to add the '__len __()' dunder method, which shows the size of the structure (and also in the other classes).

import datetime
import struct

# Volume name and Pathname of last mount point encoding
ENCODING = 'latin-1' # generally latin-1 is used

class Superblock:
    """
    Class representing the Superblock of an ext2 filesystem.
    The superblock contains all the information about the configuration of the filesystem.
    (there is a copy of the superblock in each block group in the filesystem)

    The superblock is always starting at the 1024th byte of the disk's partition,
    which normally happens to be the first byte of the 3rd sector (assuming 512-byte sectors).
    """
    def __init__(self, data=bytes(1024),
                 s_inodes_count=None, s_blocks_count=None, s_r_blocks_count=None,
                 s_free_blocks_count=None, s_free_inodes_count=None, s_first_data_block=None,
                 s_log_block_size=None, s_log_frag_size=None, s_blocks_per_group=None,
                 s_frags_per_group=None, s_inodes_per_group=None, s_mtime=None,
                 s_wtime=None, s_mnt_count=None, s_max_mnt_count=None, s_magic=None,
                 s_state=None, s_errors=None, s_minor_rev_level=None, s_lastcheck=None,
                 s_checkinterval=None, s_creator_os=None, s_rev_level=None,
                 s_def_resuid=None, s_def_resgid=None, s_first_ino=None, s_inode_size=None,
                 s_block_group_nr=None, s_feature_compat=None, s_feature_incompat=None,
                 s_feature_ro_compat=None, s_uuid=None, s_volume_name=None,
                 s_last_mounted=None, s_algorithm_usage_bitmap=None, s_prealloc_blocks=None,
                 s_prealloc_dir_blocks=None, s_padding1=None, s_reserved=None
                 ):
        
        # All fields in the superblock (as in all other ext2 structures) are stored
        # on the disk in little endian format (<)

        # 'I' represents a 32-bit unsigned int format
        # 'i' represents a 32-bit signed int format
        # See more in: https://docs.python.org/3/library/struct.html

        p_s_inodes_count      = struct.unpack("<I", data[0:4])[0]
        p_s_blocks_count      = struct.unpack("<I", data[4:8])[0]
        p_s_r_blocks_count    = struct.unpack("<I", data[8:12])[0]
        p_s_free_blocks_count = struct.unpack("<I", data[12:16])[0]
        p_s_free_inodes_count = struct.unpack("<I", data[16:20])[0]
        p_s_first_data_block  = struct.unpack("<I", data[20:24])[0]
        p_s_log_block_size    = struct.unpack("<I", data[24:28])[0]
        p_s_log_frag_size     = struct.unpack("<i", data[28:32])[0]
        p_s_blocks_per_group  = struct.unpack("<I", data[32:36])[0]
        p_s_frags_per_group   = struct.unpack("<I", data[36:40])[0]
        p_s_inodes_per_group  = struct.unpack("<I", data[40:44])[0]
        #
        p_s_mtime             = struct.unpack("<I", data[44:48])[0]
        p_s_wtime             = struct.unpack("<I", data[48:52])[0]
        p_s_mnt_count         = struct.unpack("<H", data[52:54])[0]
        p_s_max_mnt_count     = struct.unpack("<H", data[54:56])[0]
        p_s_magic             = struct.unpack("<H", data[56:58])[0]
        p_s_state             = struct.unpack("<H", data[58:60])[0]
        p_s_errors            = struct.unpack("<H", data[60:62])[0]
        p_s_minor_rev_level   = struct.unpack("<H", data[62:64])[0]
        p_s_lastcheck         = struct.unpack("<I", data[64:68])[0]
        p_s_checkinterval     = struct.unpack("<I", data[68:72])[0]
        p_s_creator_os        = struct.unpack("<I", data[72:76])[0]
        p_s_rev_level         = struct.unpack("<I", data[76:80])[0]
        p_s_def_resuid        = struct.unpack("<H", data[80:82])[0]
        p_s_def_resgid        = struct.unpack("<H", data[82:84])[0]
        #
        p_s_first_ino         = struct.unpack("<I", data[84:88])[0]
        p_s_inode_size        = struct.unpack("<H", data[88:90])[0]
        p_s_block_group_nr    = struct.unpack("<H", data[90:92])[0]
        #
        p_s_feature_compat    = struct.unpack("<I", data[92:96])[0]
        p_s_feature_incompat  = struct.unpack("<I", data[96:100])[0]
        p_s_feature_ro_compat = struct.unpack("<I", data[100:104])[0]
        p_s_uuid              = struct.unpack("<BBBBBBBBBBBBBBBB", data[104:120])
        p_s_volume_name       = data[120:136] # Volume name
        p_s_last_mounted      = data[136:200] # Pathname of last mount point
        p_s_algorithm_usage_bitmap = struct.unpack("<I", data[200:204])[0]
        p_s_prealloc_blocks        = struct.unpack("<B", data[204:205])[0]
        p_s_prealloc_dir_blocks    = struct.unpack("<B", data[205:206])[0]
        #
        p_s_padding1          = struct.unpack("<H", data[206:208])[0]
        p_s_reserved          = data[208:1024]
        # for ext3 there are some more fields: https://www.nongnu.org/ext2-doc/ext2.html#superblock

        self._raw_data = data # the raw read from the disk (the 1024 bytes of the superblock).

        self.s_inodes_count      = s_inodes_count      or p_s_inodes_count
        self.s_blocks_count      = s_blocks_count      or p_s_blocks_count
        self.s_r_blocks_count    = s_r_blocks_count    or p_s_r_blocks_count
        self.s_free_blocks_count = s_free_blocks_count or p_s_free_blocks_count
        self.s_free_inodes_count = s_free_inodes_count or p_s_free_inodes_count
        self.s_first_data_block  = s_first_data_block  or p_s_first_data_block
        self.s_log_block_size    = s_log_block_size    or p_s_log_block_size
        self.s_log_frag_size     = s_log_frag_size     or p_s_log_frag_size
        self.s_blocks_per_group  = s_blocks_per_group  or p_s_blocks_per_group
        self.s_frags_per_group   = s_frags_per_group   or p_s_frags_per_group
        self.s_inodes_per_group  = s_inodes_per_group  or p_s_inodes_per_group
        #
        self.s_mtime             = s_mtime             or p_s_mtime
        self.s_wtime             = s_wtime             or p_s_wtime
        self.s_mnt_count         = s_mnt_count         or p_s_mnt_count
        self.s_max_mnt_count     = s_max_mnt_count     or p_s_max_mnt_count
        self.s_magic             = s_magic             or p_s_magic
        self.s_state             = s_state             or p_s_state
        self.s_errors            = s_errors            or p_s_errors
        self.s_minor_rev_level   = s_minor_rev_level   or p_s_minor_rev_level
        self.s_lastcheck         = s_lastcheck         or p_s_lastcheck
        self.s_checkinterval     = s_checkinterval     or p_s_checkinterval
        self.s_creator_os        = s_creator_os        or p_s_creator_os
        self.s_rev_level         = s_rev_level         or p_s_rev_level
        self.s_def_resuid        = s_def_resuid        or p_s_def_resuid
        self.s_def_resgid        = s_def_resgid        or p_s_def_resgid
        #
        self.s_first_ino         = s_first_ino         or p_s_first_ino
        self.s_inode_size        = s_inode_size        or p_s_inode_size
        self.s_block_group_nr    = s_block_group_nr    or p_s_block_group_nr
        #
        self.s_feature_compat         = s_feature_compat         or p_s_feature_compat
        self.s_feature_incompat       = s_feature_incompat       or p_s_feature_incompat
        self.s_feature_ro_compat      = s_feature_ro_compat      or p_s_feature_ro_compat
        self.s_uuid                   = s_uuid                   or p_s_uuid
        self.s_volume_name            = s_volume_name            or p_s_volume_name
        self.s_last_mounted           = s_last_mounted           or p_s_last_mounted
        self.s_algorithm_usage_bitmap = s_algorithm_usage_bitmap or p_s_algorithm_usage_bitmap
        self.s_prealloc_blocks        = s_prealloc_blocks        or p_s_prealloc_blocks
        self.s_prealloc_dir_blocks    = s_prealloc_dir_blocks    or p_s_prealloc_dir_blocks
        #
        self.s_padding1 = s_padding1 or p_s_padding1
        self.s_reserved = s_reserved or p_s_reserved

        # 's_': superblock ; 'p_': parsed

    @property
    def raw_data(self):
        """
        Bytes to be parsed.
        (are the 1024 bytes corresponding to the structure of the superblock)
        """
        return self._raw_data

    @raw_data.setter
    def raw_data(self, value):
        pass

    # ---

    @property
    def s_inodes_count(self):
        """Total number of inodes"""
        return self._s_inodes_count

    @s_inodes_count.setter
    def s_inodes_count(self, value):
        self._s_inodes_count = int(value)

    @property
    def s_blocks_count(self):
        """Filesystem size in blocks"""
        return self._s_blocks_count

    @s_blocks_count.setter
    def s_blocks_count(self, value):
        self._s_blocks_count = int(value)

    @property
    def s_r_blocks_count(self):
        """Number of reserved blocks"""
        return self._s_r_blocks_count

    @s_r_blocks_count.setter
    def s_r_blocks_count(self, value):
        self._s_r_blocks_count = int(value)

    @property
    def s_free_blocks_count(self):
        """
        Free blocks counter
        (data blocks + directory-entry blocks)
        """
        return self._s_free_blocks_count

    @s_free_blocks_count.setter
    def s_free_blocks_count(self, value):
        self._s_free_blocks_count = int(value)

    @property
    def s_free_inodes_count(self):
        """
        Free inodes counter
        The inodes are each of the entries in the inode table, and this table
        has a calculable size. Creating an inode does not mean that a new block
        is allocated from the file system, but rather to fill in a new entry in the table.
        Therefore, "s_free_inodes_count" is an independent calculation of "s_free_blocks_count".
        """
        return self._s_free_inodes_count

    @s_free_inodes_count.setter
    def s_free_inodes_count(self, value):
        self._s_free_inodes_count = int(value)

    @property
    def s_first_data_block(self):
        """
        Number of first useful block.
        It is the id of the block containing the superblock structure.

        This value is always 0 for filesystems with a block size larger than 1KB,
        and always 1 for filesystems with a block size of 1KB, since the superblock
        is always starting at the 1024th byte of the disk's partition.
        """
        return self._s_first_data_block

    @s_first_data_block.setter
    def s_first_data_block(self, value):
        self._s_first_data_block = int(value)

    @property
    def s_log_block_size(self):
        """
        The s_log_block_size field expresses the block size as a power of 2,
        using 1,024 bytes as the unit. Thus, 0 denotes 1,024-byte blocks,
        1 denotes 2,048-byte blocks, and so on.
        """
        size = 2**(10+self._s_log_block_size)
        return size
        
        # I think this would be fine, that is, if from 'outside' someone
        # wants to see the size of the block, they will directly see the size in bytes.
        # And that the true value of the attribute (0, 1 or 2), can only be accessed internally
        # (from here, the source code, doing self._s_log_block_size).

    @s_log_block_size.setter
    def s_log_block_size(self, value):
        # Valid block size values are 1024, 2048 and 4096 bytes.
        # (actually, in Linux the block size is limited by the architecture page size)
        if value > 2:
            value = 2
        self._s_log_block_size = value

    @property
    def s_log_frag_size(self):
        """
        Fragment size (not implemented in ext2)
        The s_log_frag_size field is currently equal to s_log_block_size,
        since block fragmentation is not yet implemented.
        """
        if(self._s_log_frag_size >= 0):
            # this is equivalent to what was done in "s_log_block_size"
            fragment_size = 1024 << self._s_log_frag_size;
        else:
            fragment_size = 1024 >> -self._s_log_frag_size;
        return fragment_size

    @s_log_frag_size.setter
    def s_log_frag_size(self, value):
        # The type of the s_log_frag_size field is signed integer,
        # unlike s_log_block_size which is unsigned integer.
        if value > 2:
            value = 2
        elif value < -2:
            value = -2
        self._s_log_frag_size = value

    @property
    def s_blocks_per_group(self):
        """Number of blocks per group"""
        return self._s_blocks_per_group

    @s_blocks_per_group.setter
    def s_blocks_per_group(self, value):
        self._s_blocks_per_group = int(value)

    @property
    def s_frags_per_group(self):
        """Number of fragments per group"""
        return self._s_frags_per_group

    @s_frags_per_group.setter
    def s_frags_per_group(self, value):
        self._s_frags_per_group = int(value)

    @property
    def s_inodes_per_group(self):
        """Number of inodes per group"""
        return self._s_inodes_per_group

    @s_inodes_per_group.setter
    def s_inodes_per_group(self, value):
        self._s_inodes_per_group = int(value)

    # ---

    @property
    def s_mtime(self):
        """Time of last mount operation [POSIX time]"""
        ts = self._s_mtime if self._s_mtime.timestamp() > 0. else "Not defined"
        return ts

    @s_mtime.setter
    def s_mtime(self, value):
        # I will use DateTime objects for timestamps (this is explained in 'inode.py')
        self._s_mtime = datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)

    @property
    def s_wtime(self):
        """Time of last write operation [POSIX time]"""
        ts = self._s_wtime if self._s_wtime.timestamp() > 0. else "Not defined"
        return ts

    @s_wtime.setter
    def s_wtime(self, value):
        self._s_wtime = datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)

    @property
    def s_mnt_count(self):
        """Mount operations counter"""
        return self._s_mnt_count

    @s_mnt_count.setter
    def s_mnt_count(self, value):
        self._s_mnt_count = int(value)

    @property
    def s_max_mnt_count(self):
        """Number of mount operations before check"""
        return self._s_max_mnt_count

    @s_max_mnt_count.setter
    def s_max_mnt_count(self, value):
        self._s_max_mnt_count = int(value)

    @property
    def s_magic(self):
        """
        Magic signature (fixed to 0xEF53)
        16bit value identifying the file system as Ext2.
        """
        return hex(self._s_magic)

    @s_magic.setter
    def s_magic(self, value):
        self._s_magic = int(value)
    
    @property
    def s_state(self):
        """
        Status flag: the s_state field stores the value 0 if the filesystem
        is mounted or was not cleanly unmounted, 1 if it was cleanly unmounted,
        and 2 if it contains errors.

        I return a string representing the status.
        """
        status = "unknown"

        if self._s_state == 0:
            status = "the filesystem is mounted or was not cleanly unmounted"
        elif self._s_state == 1:
            status = "the filesystem was cleanly unmounted"
        elif self._s_state == 2:
            status = "the filesystem contains errors"

        return status

    @s_state.setter
    def s_state(self, value):
        self._s_state = int(value)

    @property
    def s_errors(self):
        """Behavior of the filesystem driver when detecting errors"""
        behavior = "undefined"

        if self._s_errors == 1:
            behavior = "continue as if nothing happened"
        elif self._s_errors == 2:
            behavior = "remount read-only"
        elif self._s_errors == 3:
            behavior = "cause a kernel panic"

        return behavior

    @s_errors.setter
    def s_errors(self, value):
        self._s_errors = int(value)

    @property
    def s_minor_rev_level(self):
        """Minor revision level"""
        return self._s_minor_rev_level

    @s_minor_rev_level.setter
    def s_minor_rev_level(self, value):
        self._s_minor_rev_level = int(value)

    @property
    def s_lastcheck(self):
        """Time of last check [POSIX time]"""
        ts = self._s_lastcheck if self._s_lastcheck.timestamp() > 0. else "Not defined"
        return ts
    
    @s_lastcheck.setter
    def s_lastcheck(self, value):
        self._s_lastcheck = datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)

    @property
    def s_checkinterval(self):
        """Time between checks [POSIX time]"""
        # https://docs.python.org/3/library/datetime.html#datetime.timedelta.total_seconds
        interval = self._s_checkinterval if self._s_checkinterval.total_seconds() > 0. else "Not defined"
        return interval

    @s_checkinterval.setter
    def s_checkinterval(self, value):
        # https://docs.python.org/3/library/datetime.html#timedelta-objects
        self._s_checkinterval = datetime.timedelta(seconds=value) # a 'timedelta' object represents a duration

    @property
    def s_creator_os(self):
        """OS where filesystem was created"""
        os = "unknown"

        if self._s_creator_os == 0:
            os = "Linux"
        elif self._s_creator_os == 1:
            os = "GNU HURD"
        elif self._s_creator_os == 2:
            os = "MASIX"
        elif self._s_creator_os == 3:
            os = "FreeBSD"
        elif self._s_creator_os == 4:
            os = "Lites"

        return os

    @s_creator_os.setter
    def s_creator_os(self, value):
        self._s_creator_os = int(value)

    @property
    def s_rev_level(self):
        """Revision level (0: revision 0 ; 1: revision 1)"""
        return self._s_rev_level

    @s_rev_level.setter
    def s_rev_level(self, value):
        self._s_rev_level = int(value)

    @property
    def s_def_resuid(self):
        """Default User ID for reserved blocks (in Linux it is 0)"""
        return self._s_def_resuid

    @s_def_resuid.setter
    def s_def_resuid(self, value):
        self._s_def_resuid = int(value)

    @property
    def s_def_resgid(self):
        """Default Group ID for reserved blocks (in Linux it is 0)"""
        return self._s_def_resgid

    @s_def_resgid.setter
    def s_def_resgid(self, value):
        self._s_def_resgid = int(value)

    # ---

    @property
    def s_first_ino(self):
        """
        Number of first nonreserved inode

        In revision 0, the first non-reserved inode is fixed to 11.
        In revision 1 and later this value may be set to any value.
        """
        return self._s_first_ino

    @s_first_ino.setter
    def s_first_ino(self, value):
        self._s_first_ino = int(value)

    @property
    def s_inode_size(self):
        """
        Size of on-disk inode structure

        In revision 0, this value is always 128.
        In revision 1 and later, this value must be a perfect power of 2
        and must be smaller or equal to the block size.
        """
        return self._s_inode_size

    @s_inode_size.setter
    def s_inode_size(self, value):
        self._s_inode_size = int(value)

    @property
    def s_block_group_nr(self):
        """Block group number of this superblock (this can be used to rebuild
        the file system from any superblock backup)"""
        return self._s_block_group_nr

    @s_block_group_nr.setter
    def s_block_group_nr(self, value):
        self._s_block_group_nr = int(value)

    # ---
    
    @property
    def s_feature_compat(self):
        """Compatible features bitmap"""
        return self._s_feature_compat

    @s_feature_compat.setter
    def s_feature_compat(self, value):
        self._s_feature_compat = value

    @property
    def s_feature_incompat(self):
        """Incompatible features bitmap"""
        return self._s_feature_incompat

    @s_feature_incompat.setter
    def s_feature_incompat(self, value):
        self._s_feature_incompat = value

    @property
    def s_feature_ro_compat(self):
        """Read-only compatible features bitmap"""
        return self._s_feature_ro_compat

    @s_feature_ro_compat.setter
    def s_feature_ro_compat(self, value):
        self._s_feature_ro_compat = value

    @property
    def s_uuid(self):
        """128-bit filesystem identifier"""
        f_id = str(self._s_uuid).replace(", ", "-")
        return f_id

    @s_uuid.setter
    def s_uuid(self, value):
        # I'm not sure if the ID is divided into 16 bytes, or is it directly a 16-byte number.
        self._s_uuid = tuple(value)

    @property
    def s_volume_name(self):
        """Volume name"""
        return self._s_volume_name

    @s_volume_name.setter
    def s_volume_name(self, value):
        # I admit both strings and bytestrings, but internally I keep bytestrings.
        if isinstance(value, str):
            value = bytes(value, ENCODING)
        self._s_volume_name = value.rstrip(b'\x00') # to remove trailing null chars

    @property
    def s_last_mounted(self):
        """Pathname of last mount point"""
        return self._s_last_mounted

    @s_last_mounted.setter
    def s_last_mounted(self, value):
        # I admit both strings and bytestrings, but internally I keep bytestrings.
        if isinstance(value, str):
            value = bytes(value, ENCODING)
        self._s_last_mounted = value.rstrip(b'\x00') # to remove trailing null chars

    @property
    def s_algorithm_usage_bitmap(self):
        """Used for compression"""
        return self._s_algorithm_usage_bitmap

    @s_algorithm_usage_bitmap.setter
    def s_algorithm_usage_bitmap(self, value):
        self._s_algorithm_usage_bitmap = value

    @property
    def s_prealloc_blocks(self):
        """Number of blocks to preallocate for regular files"""
        return self._s_prealloc_blocks

    @s_prealloc_blocks.setter
    def s_prealloc_blocks(self, value):
        self._s_prealloc_blocks = int(value)

    @property
    def s_prealloc_dir_blocks(self):
        """Number of blocks to preallocate for directories"""
        return self._s_prealloc_dir_blocks

    @s_prealloc_dir_blocks.setter
    def s_prealloc_dir_blocks(self, value):
        self._s_prealloc_dir_blocks = int(value)

    # ---

    @property
    def s_padding1(self):
        """Alignment to word [2 bytes]"""
        return self._s_padding1

    @s_padding1.setter
    def s_padding1(self, value):
        self._s_padding1 = value

    @property
    def s_reserved(self):
        """Nulls to pad out 1,024 bytes [816 bytes]"""
        return self._s_reserved

    @s_reserved.setter
    def s_reserved(self, value):
        self._s_reserved = value


    # I do not show all the fields, only the ones that I found most interesting.
    def __str__(self):
        return (
                f"Total number of inodes:                          {self.s_inodes_count}\n"
                f"Filesystem size in blocks:                       {self.s_blocks_count}\n"
                f"Number of reserved blocks:                       {self.s_r_blocks_count}\n"
                f"Free blocks counter:                             {self.s_free_blocks_count}\n"
                f"Free inodes counter:                             {self.s_free_inodes_count}\n"
                f"Number of first useful block:                    {self.s_first_data_block}\n"
                f"Block size:                                      {self.s_log_block_size}\n"
                f"Fragment size:                                   {self.s_log_frag_size}\n"
                f"Number of blocks per group:                      {self.s_blocks_per_group}\n"
                f"Number of fragments per group:                   {self.s_frags_per_group}\n"
                f"Number of inodes per group:                      {self.s_inodes_per_group}\n"
                f"Time of last mount operation:                    {self.s_mtime}\n"
                f"Time of last write operation:                    {self.s_wtime}\n"
                f"Mount operations counter:                        {self.s_mnt_count}\n"
                f"Magic signature:                                 {self.s_magic}\n"
                f"Status flag:                                     {self.s_state}\n"
                f"Time of last check:                              {self.s_lastcheck}\n"
                f"Time between checks:                             {self.s_checkinterval}\n"
                f"OS where filesystem was created:                 {self.s_creator_os}\n"
                f"Revision level:                                  {self.s_rev_level}\n"
                f"Number of first nonreserved inode:               {self.s_first_ino}\n"
                f"Size of on-disk inode structure:                 {self.s_inode_size}\n"
                f"Block group number of this superblock:           {self.s_block_group_nr}\n"
                f"Filesystem identifier:                           {self.s_uuid}\n"
                f"Volume name:                                     {self.s_volume_name.decode(ENCODING)}\n"
                f"Pathname of last mount point:                    {self.s_last_mounted.decode(ENCODING)}\n"
                f"Number of blocks to preallocate:                 {self.s_prealloc_blocks}\n"
                f"Number of blocks to preallocate for directories: {self.s_prealloc_dir_blocks}\n"
            )

    
    # Ext2 Superblock Operations

    def read_inode():
        pass

    def write_inode():
        pass

    def put_inode():
        pass

    def delete_inode():
        pass

    def put_super():
        pass

    def write_super():
        pass

    def statfs():
        pass

    def remount_fs():
        pass
    
